from . import MenuItem


class SubmenuItem(MenuItem):
    """
    A menu item to open a submenu
    """

    def __init__(self, text, submenu, menu=None, should_exit=False):
        """
        :ivar ConsoleMenu self.submenu: The submenu to be opened when this item is selected
        """
        super(SubmenuItem, self).__init__(text=text, menu=menu, should_exit=should_exit)

        self.submenu = submenu
        if menu:
            self.get_submenu().parent = menu

    def set_menu(self, menu):
        """
        Sets the menu of this item.
        Should be used instead of directly accessing the menu attribute for this class.

        :param ConsoleMenu menu: the menu
        """
        self.menu = menu
        self.get_submenu().parent = menu

    def set_up(self):
        """
        This class overrides this method
        """
        self.menu.pause()
        self.menu.clear_screen()

    def action(self):
        """
        This class overrides this method
        """
        self.get_submenu().start()

    def clean_up(self):
        """
        This class overrides this method
        """
        self.get_submenu().join()
        self.menu.clear_screen()
        self.menu.resume()

    def get_return(self):
        """
        :return: The returned value in the submenu
        """
        return self.get_submenu().returned_value

    def get_submenu(self):
        """
        We unwrap the submenu variable in case it is a reference to a method that returns a submenu
        """
        return self.submenu if not callable(self.submenu) else self.submenu()



from typing import Callable, Any

import logging

from . import utils
from .auth.PrivilegeHandler import PrivilegeHandler


def privileged(method: Callable) -> Any:
    """Decorator function to execute a method with proper privileges. They are then dropped when method is executed."""

    def exec_with_privileges(*args, **kw):
        logging.debug(f'Executing method `{method.__name__}` with raised privileges...')
        utils.exec_by_platform(PrivilegeHandler.get_instance().raise_privileges, lambda: None, lambda: None)
        result = method(*args, **kw)
        utils.exec_by_platform(PrivilegeHandler.get_instance().drop_privileges, lambda: None, lambda: None)

        return result

    return exec_with_privileges



import logging
import re
import shlex
import sys
import tarfile
import tempfile
import time
from itertools import islice
from multiprocessing.dummy import Pool
from typing import List, Dict, Generator, Optional, Set, Tuple, Union, Any

import chardet
import docker.models.containers
from docker import DockerClient
from docker.errors import APIError
from docker.types import Ulimit
from docker.utils import version_lt, version_gte

from .DockerImage import DockerImage
from .exec_stream.DockerExecStream import DockerExecStream
from .stats.DockerMachineStats import DockerMachineStats
from ... import utils
from ...decorators import privileged
from ...event.EventDispatcher import EventDispatcher
from ...exceptions import MountDeniedError, MachineAlreadyExistsError, DockerPluginError, \
    MachineBinaryError, MachineNotRunningError, PrivilegeError, InvocationError
from ...model.Interface import Interface
from ...model.Lab import Lab
from ...model.Link import Link, BRIDGE_LINK_NAME
from ...model.Machine import Machine, MACHINE_CAPABILITIES
from ...setting.Setting import Setting
from ...utils import parse_docker_engine_version

RP_FILTER_NAMESPACE = "net.ipv4.conf.%s.rp_filter"
OCI_RUNTIME_RE = re.compile(
    r"OCI runtime exec failed(.*?)(stat (.*): no such file or directory|exec: \"(.*)\": executable file not found)"
)

IFACE_SYSCTL_RE = re.compile(r"net\.ipv[4,6]\.(conf|neigh)\.eth\d+")

# Known commands that each container should execute
# Run order: shared.startup, machine.startup and machine.meta['exec_commands']
STARTUP_COMMANDS = [
    # Unmount the /etc/resolv.conf and /etc/hosts files, automatically mounted by Docker inside the container.
    # In this way, they can be overwritten by custom user files.
    "umount /etc/resolv.conf",
    "umount /etc/hosts",

    # Copy the machine folder (if present) from the hostlab directory into the root folder of the container
    # In this way, files are all replaced in the container root folder
    "if [ -d \"/hostlab/{machine_name}\" ]; then "
    "(cd /hostlab/{machine_name} && tar c .) | (cd / && tar xhf - --no-same-owner --no-same-permissions); fi",

    # If /etc/hosts is not configured by the user, add the default mappings
    "if [ ! -s \"/etc/hosts\" ]; then "
    "echo '127.0.0.1 localhost' > /etc/hosts",
    "echo '::1 localhost' >> /etc/hosts",
    "echo '127.0.1.1 {machine_name}' >> /etc/hosts",
    "fi",

    # Give proper permissions to /var/www
    "if [ -d \"/var/www\" ]; then "
    "chmod -R 777 /var/www/*; fi",

    # Give proper permissions to Quagga files (if present)
    "if [ -d \"/etc/quagga\" ]; then "
    "chown -R quagga:quagga /etc/quagga/",
    "chmod 640 /etc/quagga/*; fi",

    # Give proper permissions to FRR files (if present)
    "if [ -d \"/etc/frr\" ]; then "
    "chown -R frr:frr /etc/frr/",
    "chmod 640 /etc/frr/*; fi",

    # If shared.startup file is present
    "if [ -f \"/hostlab/shared.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/shared.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/shared.startup",
    "/hostlab/shared.startup &> /var/log/shared.log; fi",

    # If .startup file is present
    "if [ -f \"/hostlab/{machine_name}.startup\" ]; then "
    # Give execute permissions to the file and execute it
    # We redirect the output "&>" to a debugging file
    "chmod u+x /hostlab/{machine_name}.startup",
    # Adds a line to enable command output
    "sed -i \"1s;^;set -x\\n\\n;\" /hostlab/{machine_name}.startup",
    "/hostlab/{machine_name}.startup &> /var/log/startup.log; fi",

    # Placeholder for user commands
    "{machine_commands}",

    "touch /tmp/EOS"
]

SHUTDOWN_COMMANDS = [
    # If machine.shutdown file is present
    "if [ -f \"/hostlab/{machine_name}.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/{machine_name}.shutdown; /hostlab/{machine_name}.shutdown; fi",

    # If shared.shutdown file is present
    "if [ -f \"/hostlab/shared.shutdown\" ]; then "
    # Give execute permissions to the file and execute it
    "chmod u+x /hostlab/shared.shutdown; /hostlab/shared.shutdown; fi"
]


class DockerMachine(object):
    """The class responsible for deploying Kathara devices as Docker container and interact with them."""
    __slots__ = ['client', '_engine_version', 'docker_image']

    def __init__(self, client: DockerClient, docker_image: DockerImage) -> None:
        self.client: DockerClient = client
        self._engine_version: str = parse_docker_engine_version(client.version()['Version'])
        self.docker_image: DockerImage = docker_image

    def deploy_machines(self, lab: Lab, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
        """Deploy all the network scenario devices as Docker containers.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.
            selected_machines (Set[str]): A set containing the name of the devices to deploy.
            excluded_machines (Set[str]): A set containing the name of the devices to exclude.

        Returns:
            None

        Raises:
            PrivilegeError: If the privileged mode is active and the user does not have root privileges.
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")

        machines = lab.machines.items()
        if selected_machines:
            machines = {
                k: v for k, v in machines if k in selected_machines
            }.items()
        elif excluded_machines:
            machines = {
                k: v for k, v in machines if k not in excluded_machines
            }.items()

        # Check and pulling machine images
        lab_images = set(map(lambda x: x[1].get_image(), machines))
        self.docker_image.check_from_list(lab_images)

        policy = Setting.get_instance().volume_mount_policy
        lab.add_option("_mount_volumes", policy in ['Prompt', 'Always'])
        machines_with_volumes = dict(filter(lambda x: len(x[1].meta['volumes']) > 0, machines))
        if len(machines_with_volumes) > 0:
            EventDispatcher.get_instance().dispatch(
                "machines_with_volumes", lab=lab, machines_with_volumes=machines_with_volumes
            )

        shared_mount = lab.general_options['shared_mount'] if 'shared_mount' in lab.general_options \
            else Setting.get_instance().shared_mount
        if shared_mount:
            if Setting.get_instance().remote_url is not None:
                logging.warning("Shared folder cannot be mounted with a remote Docker connection.")
            else:
                lab.create_shared_folder()

        EventDispatcher.get_instance().dispatch("machines_deploy_started", items=machines)

        # Deploy all lab machines.
        # If there is no lab.dep file, machines can be deployed using multithreading.
        # If not, they're started sequentially
        if not lab.has_dependencies:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(machines, pool_size)

            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=self._deploy_and_start_machine, iterable=chunk)
        else:
            for item in machines:
                self._deploy_and_start_machine(item)

        EventDispatcher.get_instance().dispatch("machines_deploy_ended")

        # Delete to avoid keeping dirty state
        del lab.general_options['_mount_volumes']

    def _deploy_and_start_machine(self, machine_item: Tuple[str, Machine]) -> None:
        """Deploy and start a Docker container from the device contained in machine_item.

        Args:
            machine_item (Tuple[str, Machine]): A tuple composed by the name of the device and a device object

        Returns:
            None
        """
        (_, machine) = machine_item

        self.create(machine)
        self.start(machine)

        EventDispatcher.get_instance().dispatch("machine_deployed", item=machine)

    def create(self, machine: Machine) -> None:
        """Create a Docker container representing the device and assign it to machine.api_object.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            None

        Raises:
            MachineAlreadyExistsError: If a device with the name specified already exists.
            APIError: If the Docker APIs return an error.
        """
        logging.debug("Creating device `%s`..." % machine.name)

        containers = self.get_machines_api_objects_by_filters(machine_name=machine.name, lab_hash=machine.lab.hash,
                                                              user=utils.get_current_user_name())
        if containers:
            raise MachineAlreadyExistsError(machine.name)

        image = machine.get_image()
        memory = machine.get_mem()
        cpus = machine.get_cpu(multiplier=1000000000)
        ulimits = [Ulimit(name=k, soft=v["soft"], hard=v["hard"]) for k, v in machine.get_ulimits().items()]

        ports_info = machine.get_ports()
        ports = None
        if ports_info:
            ports = {}
            for (host_port, protocol), guest_port in ports_info.items():
                ports['%d/%s' % (guest_port, protocol)] = host_port

        # Get the global machine metadata into a local variable (just to avoid accessing the lab object every time)
        global_machine_metadata = machine.lab.global_machine_metadata

        # If bridged is required in command line but not defined in machine meta, add it.
        if "bridged" in global_machine_metadata and not machine.is_bridged():
            machine.add_meta("bridged", True)

        if ports and not machine.is_bridged():
            logging.warning(
                "To expose ports of device `%s` on the host, "
                "you have to specify the `bridged` option on that device." % machine.name
            )

        # If any exec command is passed in command line, add it.
        if "exec" in global_machine_metadata:
            machine.add_meta("exec", global_machine_metadata["exec"])

        # Get the first network object, if defined.
        # This should be used in container create function
        first_network = None
        first_machine_iface = None
        if machine.interfaces:
            first_machine_iface = machine.interfaces[0]
            first_network = first_machine_iface.link.api_object

        if machine.is_bridged():
            machine.add_meta("bridged_iface", max(machine.interfaces.keys()) + 1 if machine.interfaces else 0)

        # If no interfaces are declared in machine, but bridged mode is required, get bridge as first link.
        # Flag that bridged is already connected (because there's another check in `start`).
        if first_machine_iface is None and machine.is_bridged():
            first_network = machine.lab.get_or_new_link(BRIDGE_LINK_NAME).api_object
            machine.add_meta("_bridge_connected", True)

        # Sysctl params to pass to the container creation
        sysctl_parameters = {RP_FILTER_NAMESPACE % x: 0 for x in ["all", "default", "lo"]}
        sysctl_parameters["net.ipv4.ip_forward"] = 1
        sysctl_parameters["net.ipv4.icmp_ratelimit"] = 0

        sysctl_first_interface = {}
        if first_machine_iface:
            if version_lt(self._engine_version, "26.0.0"):
                sysctl_first_interface = {RP_FILTER_NAMESPACE % "eth0": 0}

        if machine.is_ipv6_enabled():
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 1
            sysctl_parameters["net.ipv6.conf.all.accept_ra"] = 0
            sysctl_parameters["net.ipv6.icmp.ratelimit"] = 0
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 0
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 0
        else:
            sysctl_parameters["net.ipv6.conf.default.disable_ipv6"] = 1
            sysctl_parameters["net.ipv6.conf.all.disable_ipv6"] = 1
            sysctl_parameters["net.ipv6.conf.default.forwarding"] = 0
            sysctl_parameters["net.ipv6.conf.all.forwarding"] = 0

        # Merge machine sysctls
        sysctl_parameters = {**sysctl_parameters, **machine.meta['sysctls'], **sysctl_first_interface}
        # Remove interface-related sysctls on newer Docker Engine versions (these are added as driver_opts)
        if version_gte(self._engine_version, "27.0.0"):
            sysctl_parameters = dict(filter(lambda x: not IFACE_SYSCTL_RE.match(x[0]), sysctl_parameters.items()))

        volumes = {}

        lab_options = machine.lab.general_options
        shared_mount = ['shared_mount'] if 'shared_mount' in lab_options else Setting.get_instance().shared_mount
        if shared_mount and machine.lab.shared_path:
            volumes[machine.lab.shared_path] = {'bind': '/shared', 'mode': 'rw'}

        # Mount the host home only if specified in settings.
        hosthome_mount = lab_options['hosthome_mount'] if 'hosthome_mount' in lab_options else \
            Setting.get_instance().hosthome_mount
        if hosthome_mount and Setting.get_instance().remote_url is None:
            volumes[utils.get_current_user_home()] = {'bind': '/hosthome', 'mode': 'rw'}

        try:
            for host_path, volume in machine.get_volumes().items():
                missing_permissions = utils.check_directory_permissions(host_path, volume['mode'])

                if not missing_permissions:
                    volumes[host_path] = {'bind': volume['guest_path'], 'mode': volume['mode']}
                else:
                    raise PermissionError(
                        f"To mount volume `{host_path}` in `{volume['guest_path']}` "
                        f"you miss the following permissions: `{', '.join(missing_permissions)}`."
                    )
        except MountDeniedError:
            logging.warning(f"Volumes of device `{machine.name}` will not be mounted.")

        privileged = machine.is_privileged()
        if privileged and not utils.is_admin():
            raise PrivilegeError(f"You must be root in order to start device `{machine.name}` in privileged mode.")
        if Setting.get_instance().remote_url is not None and privileged:
            privileged = False
            logging.warning("Privileged flag is ignored with a remote Docker connection.")

        networking_config = None
        if first_machine_iface:
            driver_opt = self._create_driver_opt(
                machine, first_machine_iface, self._get_iface_sysctls(machine.get_sysctls(), first_machine_iface.num)
            )

            networking_config = {
                first_network.name: self.client.api.create_endpoint_config(driver_opt=driver_opt)
            }

        container_name = self.get_container_name(machine.name, machine.lab.hash)

        try:
            labels = {"name": machine.name,
                      "lab_hash": machine.lab.hash,
                      "user": utils.get_current_user_name(),
                      "app": "kathara",
                      "shell": machine.meta["shell"]
                      if "shell" in machine.meta
                      else Setting.get_instance().device_shell
                      }
            if machine.is_bridged():
                labels["bridged_iface"] = str(machine.meta["bridged_iface"])

            entrypoint = shlex.split(machine.meta["entrypoint"]) if "entrypoint" in machine.meta else None
            args = machine.meta["args"] if "args" in machine.meta and machine.meta["args"] else None
            if args:
                args = shlex.split(args) if type(args) == str else args

            machine_container = self.client.containers.create(image=image,
                                                              name=container_name,
                                                              hostname=machine.name,
                                                              cap_add=MACHINE_CAPABILITIES if not privileged else None,
                                                              privileged=privileged,
                                                              network=first_network.name if first_network else None,
                                                              network_mode="bridge" if first_network else "none",
                                                              networking_config=networking_config,
                                                              environment=machine.meta['envs'],
                                                              sysctls=sysctl_parameters,
                                                              mem_limit=memory,
                                                              nano_cpus=cpus,
                                                              ports=ports,
                                                              tty=True,
                                                              stdin_open=True,
                                                              detach=True,
                                                              volumes=volumes,
                                                              labels=labels,
                                                              ulimits=ulimits,
                                                              entrypoint=entrypoint,
                                                              command=args
                                                              )
        except APIError as e:
            raise e

        # Pack machine files into a tar.gz and extract its content inside `/`
        tar_data = machine.pack_data()
        if tar_data:
            self.copy_files(machine_container, "/", tar_data)

        machine.api_object = machine_container

    def connect_interface(self, machine: Machine, interface: Interface) -> None:
        """Connect the Docker container representing the machine to a specified collision domain.

        Args:
            machine (Kathara.model.Machine.Machine): A Kathara device.
            interface (Kathara.model.Interface.Interface): A Kathara interface object.

        Returns:
            None

        Raises:
            DockerPluginError: If Kathara has been left in an inconsistent state.
            APIError: If the Docker APIs return an error.
        """
        attached_networks = machine.api_object.attrs["NetworkSettings"]["Networks"]

        if interface.link.api_object.name not in attached_networks:
            driver_opt = self._create_driver_opt(
                machine, interface, self._get_iface_sysctls(machine.get_sysctls(), interface.num)
            )

            try:
                interface.link.api_object.connect(
                    machine.api_object,
                    driver_opt=driver_opt
                )
            except APIError as e:
                if e.response.status_code == 500 and \
                        ("network does not exist" in e.explanation or "endpoint does not exist" in e.explanation):
                    raise DockerPluginError(
                        "Kathara has been left in an inconsistent state! Please run `kathara wipe`.")
                else:
                    raise e

    @staticmethod
    def _get_iface_sysctls(sysctls: Dict[str, int], interface_num: int) -> Set[str]:
        """Extract and transform sysctl settings specific to a particular network interface.

        Args:
            sysctls (Dict[str, int]): The sysctls for the device.
            interface_num (int): The interface number for which to match and transform sysctl settings.

        Returns:
            Set[str]: A set containing transformed sysctl settings strings for the specified interface.
                Each string includes the setting in "key=value" format with the interface name replaced by 'IFNAME'.
        """
        iface_sysctls = set()
        for sysctl_name, value in sysctls.items():
            if re.match(rf"net\.ipv[4,6]\.(conf|neigh)\.eth{interface_num}", sysctl_name):
                sysctl = f"{re.sub(rf'eth{interface_num}', 'IFNAME', sysctl_name)}={value}"
                iface_sysctls.add(sysctl)
        return iface_sysctls

    def _create_driver_opt(self, machine: Machine, interface: Interface, iface_sysctls: Set[str]) -> Dict[str, str]:
        """Create a dict containing the network driver options for a device.

        Args:
            machine (Kathara.model.Machine.Machine): The Kathara device to be attached to the interface.
            interface (Kathara.model.Interface.Interface): The interface to be attached to the device.
            iface_sysctls (Set[str]): Sysctls in Docker driver_opt format for the interface.

        Returns:
            Dict[str, str]: A dict containing the network driver options for a device.
        """
        driver_opt = {'kathara.iface': str(interface.num), 'kathara.link': interface.link.name}
        if version_gte(self._engine_version, "27.0.0"):
            sysctl_opts = {"net.ipv4.conf.IFNAME.rp_filter=0"}

            if machine.is_ipv6_enabled():
                sysctl_opts.update({"net.ipv6.conf.IFNAME.disable_ipv6=0", "net.ipv6.conf.IFNAME.forwarding=1"})
            else:
                sysctl_opts.add("net.ipv6.conf.IFNAME.disable_ipv6=1")

            sysctl_opts.update(iface_sysctls)

            driver_opt["com.docker.network.endpoint.sysctls"] = ",".join(sysctl_opts)

        if interface.mac_address:
            driver_opt['kathara.mac_addr'] = interface.mac_address

        return driver_opt

    @staticmethod
    def disconnect_from_link(machine: Machine, link: Link) -> None:
        """Disconnect the Docker container representing the machine from a specified collision domain.

        Args:
            machine (Kathara.model.Machine): A Kathara device.
            link (Kathara.model.Link): A Kathara collision domain object.

        Returns:
            None
        """
        attached_networks = machine.api_object.attrs["NetworkSettings"]["Networks"]

        if link.api_object.name in attached_networks:
            link.api_object.disconnect(machine.api_object)

    def start(self, machine: Machine) -> None:
        """Start the Docker container representing the device.

        Connect the container to the networks, run the startup commands and open a terminal (if requested).

        Args:
           machine (Kathara.model.Machine.Machine): A Kathara device.

        Returns:
            None

        Raises:
            MountDeniedError: If the host drive is not shared with Docker.
            DockerPluginError: If Kathara has been left in an inconsistent state.
            APIError: If the Docker APIs return an error.
        """
        logging.debug("Starting device `%s`..." % machine.name)

        try:
            machine.api_object.start()
        except APIError as e:
            if e.response.status_code == 500 and e.explanation.startswith('Mounts denied'):
                raise MountDeniedError("Host drive is not shared with Docker.")
            elif e.response.status_code == 500 and \
                    ("network does not exist" in e.explanation or "endpoint does not exist" in e.explanation):
                raise DockerPluginError("Kathara has been left in an inconsistent state! Please run `kathara wipe`.")
            else:
                raise e

        # Connect the container to its networks (starting from the second, the first is already connected in `create`)
        # This should be done after the container start because Docker causes a non-deterministic order when attaching
        # networks before container startup.
        for (iface_num, machine_iface) in islice(machine.interfaces.items(), 1, None):
            logging.debug(
                f"Connecting device `{machine.name}` to collision domain `{machine_iface.link.name}` "
                f"on interface {iface_num}..."
            )
            self.connect_interface(machine, machine_iface)

        # Bridged connection required but not added in `deploy` method.
        if "_bridge_connected" not in machine.meta and machine.is_bridged():
            bridge_link = machine.lab.get_or_new_link(BRIDGE_LINK_NAME).api_object
            bridge_link.connect(machine.api_object)

        # Append executed machine startup commands inside the /var/log/startup.log file
        if machine.meta['exec_commands']:
            new_commands = []
            for command in machine.meta['exec_commands']:
                new_commands.append("echo \"++ %s\" &>> /var/log/startup.log" % command)
                new_commands.append(command)
            machine.meta['exec_commands'] = new_commands

        # Build the final startup commands string
        startup_commands_string = "; ".join(STARTUP_COMMANDS).format(
            machine_name=machine.name,
            machine_commands="; ".join(machine.meta['exec_commands']) if machine.meta['exec_commands'] else ":"
        )

        logging.debug(f"Executing startup command on `{machine.name}`: {startup_commands_string}")

        try:
            # Execute the startup commands inside the container (without privileged flag so basic permissions are used)
            self._exec_run(machine.api_object,
                           cmd=[machine.api_object.labels['shell'], '-c', startup_commands_string],
                           stdout=True,
                           stderr=True,
                           privileged=False,
                           detach=True
                           )
        except MachineBinaryError as e:
            machine.add_meta('num_terms', 0)

            logging.warning(f"Shell `{e.binary}` not found in "
                            f"image `{machine.get_image()}` of device `{machine.name}`. "
                            f"Startup commands will not be executed and terminal will not open. "
                            f"Please specify a valid shell for this device."
                            )

        machine.api_object.reload()

        # Delete to avoid keeping dirty state
        if '_bridge_connected' in machine.meta:
            del machine.meta['_bridge_connected']

    def undeploy(self, lab_hash: str, selected_machines: Set[str] = None, excluded_machines: Set[str] = None) -> None:
        """Undeploy the devices contained in the network scenario defined by the lab_hash.

        If a set of selected_machines is specified, undeploy only the specified devices.

        Args:
            lab_hash (str): The hash of the network scenario to undeploy.
            selected_machines (Set[str]): A set containing the name of the devices to undeploy.
            excluded_machines (Set[str]): A set containing the name of the devices to exclude.

        Returns:
            None

        Raises:
            InvocationError: If both `selected_machines` and `excluded_machines` are specified.
        """
        if selected_machines and excluded_machines:
            raise InvocationError(f"You can either specify `selected_machines` or `excluded_machines`.")

        containers = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, user=utils.get_current_user_name())
        if selected_machines:
            containers = [item for item in containers if item.labels["name"] in selected_machines]
        elif excluded_machines:
            containers = [item for item in containers if item.labels["name"] not in excluded_machines]

        if len(containers) > 0:
            pool_size = utils.get_pool_size()
            items = utils.chunk_list(containers, pool_size)

            EventDispatcher.get_instance().dispatch("machines_undeploy_started", items=containers)

            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=self._undeploy_machine, iterable=chunk)

            EventDispatcher.get_instance().dispatch("machines_undeploy_ended")

    def wipe(self, user: str = None) -> None:
        """Undeploy all the running devices of the specified user. If user is None, it undeploy all the running devices.

        Args:
            user (str): The name of a current user on the host.

        Returns:
            None
        """
        containers = self.get_machines_api_objects_by_filters(user=user)

        pool_size = utils.get_pool_size()
        items = utils.chunk_list(containers, pool_size)

        with Pool(pool_size) as machines_pool:
            for chunk in items:
                machines_pool.map(func=self._undeploy_machine, iterable=chunk)

    def _undeploy_machine(self, machine_api_object: docker.models.containers.Container) -> None:
        """Undeploy a Docker container.

        Args:
            machine_api_object (docker.models.containers.Container): The Docker container to undeploy.

        Returns:
            None
        """
        self._delete_machine(machine_api_object)

        EventDispatcher.get_instance().dispatch("machine_undeployed", item=machine_api_object)

    def connect(self, lab_hash: str, machine_name: str, user: str = None, shell: str = None,
                logs: bool = False, wait: Union[bool, Tuple[int, float]] = True) -> None:
        """Open a stream to the Docker container specified by machine_name using the specified shell.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device to connect.
            user (str): The name of a current user on the host.
            shell (str): The path to the desired shell.
            logs (bool): If True, print the logs of the startup command.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before giving control to the user. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to wait
                for each retry. Default is True.

        Returns:
            None

        Raises:
            MachineNotRunningError: If the specified device is not running.
            ValueError: If the wait values is neither a boolean nor a tuple, or an invalid tuple.
        """
        containers = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name, user=user)
        if not containers:
            raise MachineNotRunningError(machine_name)
        container = containers.pop()

        if not shell:
            shell = shlex.split(container.labels['shell'])
        else:
            shell = shlex.split(shell)

        logging.debug("Connect to device `%s` with shell: %s" % (machine_name, shell))

        if isinstance(wait, tuple):
            if len(wait) != 2:
                raise ValueError("Invalid `wait` value.")

            n_retries, retry_interval = wait
            should_wait = True
        elif isinstance(wait, bool):
            n_retries = None
            retry_interval = 1
            should_wait = wait
        else:
            raise ValueError("Invalid `wait` value.")

        startup_waited = 0
        if should_wait:
            startup_waited = self._wait_startup_execution(container, n_retries=n_retries, retry_interval=retry_interval)

            EventDispatcher.get_instance().dispatch("machine_startup_wait_ended")

            if startup_waited == 2:
                return

        if logs and Setting.get_instance().print_startup_log:
            # Get the logs, if the command fails it means that the shell is not found.
            cat_logs_cmd = "cat /var/log/shared.log /var/log/startup.log /var/kathara/*"
            startup_command = [item for item in shell]
            startup_command.extend(['-c', cat_logs_cmd])
            exec_result = self._exec_run(container,
                                         cmd=startup_command,
                                         stdout=True,
                                         stderr=False,
                                         privileged=False,
                                         detach=False
                                         )
            char_encoding = chardet.detect(exec_result['output']) if exec_result['output'] else None
            startup_output = exec_result['output'].decode(char_encoding['encoding']) if exec_result['output'] else None

            if startup_output:
                sys.stdout.write("--- Startup Commands Log\n")
                sys.stdout.write(startup_output)
                sys.stdout.write("--- End Startup Commands Log\n")

                if not startup_waited:
                    sys.stdout.write("!!! Executing other commands in background !!!\n")

                sys.stdout.flush()

        resp = self.client.api.exec_create(container.id,
                                           shell,
                                           stdout=True,
                                           stderr=True,
                                           stdin=True,
                                           tty=True,
                                           privileged=False
                                           )

        exec_output = self.client.api.exec_start(resp['Id'],
                                                 tty=True,
                                                 socket=True
                                                 )

        def tty_connect():
            from .terminal.DockerTTYTerminal import DockerTTYTerminal
            DockerTTYTerminal(exec_output, self.client, resp['Id']).start()

        def cmd_connect():
            from .terminal.DockerNPipeTerminal import DockerNPipeTerminal
            DockerNPipeTerminal(exec_output, self.client, resp['Id']).start()

        utils.exec_by_platform(tty_connect, cmd_connect, tty_connect)

    def exec(self, lab_hash: str, machine_name: str, command: Union[str, List], user: str = None,
             tty: bool = True, wait: Union[bool, Tuple[int, float]] = False, stream: bool = True) \
            -> Union[DockerExecStream, Tuple[bytes, bytes, int]]:
        """Execute the command on the Docker container specified by the lab_hash and the machine_name.

        Args:
            lab_hash (str): The hash of the network scenario containing the device.
            machine_name (str): The name of the device.
            user (str): The name of a current user on the host.
            command (Union[str, List]): The command to execute.
            tty (bool): If True, open a new tty.
            wait (Union[bool, Tuple[int, float]]): If True, wait indefinitely until the end of the startup commands
                execution before executing the command. If a tuple is provided, the first value indicates the
                number of retries before stopping waiting and the second value indicates the time interval to
                wait for each retry. Default is False.
            stream (bool): If True, return a DockerExecStream object.
                If False, returns a tuple containing the complete stdout, the stderr, and the return code of the command.

        Returns:
             Union[DockerExecStream, Tuple[bytes, bytes, int]]: A DockerExecStream object or
             a tuple containing the stdout, the stderr and the return code of the command.

        Raises:
            MachineNotRunningError: If the specified device is not running.
            MachineBinaryError: If the binary of the command is not found.
            ValueError: If the wait values is neither a boolean nor a tuple, or an invalid tuple.
        """
        logging.debug("Executing command `%s` to device with name: %s" % (command, machine_name))

        containers = self.get_machines_api_objects_by_filters(lab_hash=lab_hash, machine_name=machine_name, user=user)
        if not containers:
            raise MachineNotRunningError(machine_name)
        container = containers.pop()

        if isinstance(wait, tuple):
            if len(wait) != 2:
                raise ValueError("Invalid `wait` value.")

            n_retries, retry_interval = wait
            should_wait = True
        elif isinstance(wait, bool):
            n_retries = None
            retry_interval = 1
            should_wait = wait
        else:
            raise ValueError("Invalid `wait` value.")

        if should_wait:
            startup_waited = self._wait_startup_execution(container, n_retries=n_retries, retry_interval=retry_interval)

            if startup_waited == 2:
                raise MachineNotRunningError(machine_name)

        command = shlex.split(command) if type(command) is str else command
        exec_result = self._exec_run(container,
                                     cmd=command,
                                     stdout=True,
                                     stderr=True,
                                     tty=tty,
                                     privileged=False,
                                     stream=stream,
                                     demux=True,
                                     detach=False
                                     )

        if stream:
            return DockerExecStream(exec_result['output'], exec_result['Id'], self.client)

        return exec_result['output'][0], exec_result['output'][1], exec_result['exit_code']

    def _exec_run(self, container: docker.models.containers.Container,
                  cmd: Union[str, List], stdout=True, stderr=True, stdin=False, tty=False,
                  privileged=False, user='', detach=False, stream=False,
                  socket=False, environment=None, workdir=None,
                  demux=False) -> Dict[str, Optional[Any]]:
        """Custom implementation of the `exec_run` method that also checks if the executed binary exists.

        Args:
            container (docker.models.containers.Container): Container object on which the command is executed.
            cmd (str or list): Command to be executed
            stdout (bool): Attach to stdout. Default: ``True``
            stderr (bool): Attach to stderr. Default: ``True``
            stdin (bool): Attach to stdin. Default: ``False``
            tty (bool): Allocate a pseudo-TTY. Default: False
            privileged (bool): Run as privileged.
            user (str): User to execute command as. Default: root
            detach (bool): If true, detach from the exec command. Default: False
            stream (bool): Stream response data. Default: False
            socket (bool): Return the connection socket to allow custom read/write operations. Default: False
            environment (dict or list): A dictionary or a list of strings in the following format
                ``["PASSWORD=xxx"]`` or ``{"PASSWORD": "xxx"}``.
            workdir (str): Path to working directory for this exec session
            demux (bool): Return stdout and stderr separately

        Returns:
            (Dict): A dict of (exit_code, Id, output)
                exit_code: (int):
                    Exit code for the executed command or ``None`` if
                    either ``stream`` or ``socket`` is ``True``.
                Id: (str):
                    The exec id from Docker.
                output: (generator, bytes, or tuple):
                    If ``stream=True``, a generator yielding response chunks.
                    If ``socket=True``, a socket object for the connection.
                    If ``demux=True``, a tuple of two bytes: stdout and stderr.
                    A bytestring containing response data otherwise.

        Raises:
            APIError: If the server returns an error.
            MachineBinaryError: If the binary of the command is not found.
        """
        resp = self.client.api.exec_create(
            container.id, cmd, stdout=stdout, stderr=stderr, stdin=stdin, tty=tty,
            privileged=privileged, user=user, environment=environment,
            workdir=workdir,
        )

        try:
            exec_output = self.client.api.exec_start(
                resp['Id'], detach=detach, tty=tty, stream=stream, socket=socket, demux=demux
            )
        except APIError as e:
            matches = OCI_RUNTIME_RE.search(e.explanation)
            if matches:
                raise MachineBinaryError(matches.group(3) or matches.group(4), container.labels['name'])

            raise e

        exit_code = self.client.api.exec_inspect(resp['Id'])['ExitCode']
        if not socket and not stream and (exit_code is not None and exit_code != 0):
            (stdout_out, _) = exec_output if demux else (exec_output, None)
            exec_stdout = ""
            if stdout_out:
                if type(stdout_out) is bytes:
                    char_encoding = chardet.detect(stdout_out)
                    exec_stdout = stdout_out.decode(char_encoding['encoding'])
                else:
                    exec_stdout = stdout_out
            matches = OCI_RUNTIME_RE.search(exec_stdout)
            if matches:
                raise MachineBinaryError(matches.group(3) or matches.group(4), container.labels['name'])

        if socket or stream:
            return {'exit_code': None, 'Id': resp['Id'], 'output': exec_output}

        return {'exit_code': int(exit_code) if exit_code is not None else None, 'Id': resp['Id'], 'output': exec_output}

    def _wait_startup_execution(self, container: docker.models.containers.Container,
                                n_retries: Optional[int] = None, retry_interval: float = 1) -> int:
        """Wait until the startup commands are executed or until the user requests the control over the device.

        Args:
            container (docker.models.containers.Container): The Docker container to wait.
            n_retries (Optional[int]): Number of retries before stopping waiting. Default is None, waits indefinitely.
            retry_interval (float): The time interval in seconds to wait for each retry. Default is 1.

        Returns:
            bool: 0 if the user requests the control before the ending of the startup. 1, if the startup ends.
                2 if an APIError occurred during execution.
        """
        logging.debug(f"Waiting startup commands execution for device {container.labels['name']}...")

        n_retries = n_retries if n_retries is None or n_retries >= 0 else abs(n_retries)
        retry_interval = retry_interval if retry_interval >= 0 else 1

        retries = 0
        is_cmd_success = False
        startup_waited = 1
        printed = False
        while not is_cmd_success:
            try:
                exec_result = self._exec_run(container,
                                             cmd="cat /tmp/EOS",
                                             stdout=True,
                                             stderr=False,
                                             privileged=False,
                                             detach=False
                                             )
                is_cmd_success = exec_result['exit_code'] == 0

                if not printed and not is_cmd_success:
                    EventDispatcher.get_instance().dispatch("machine_startup_wait_started")
                    printed = True

                # If the user requests the control, break the while loop
                if utils.exec_by_platform(utils.wait_user_input_linux,
                                          utils.wait_user_input_windows,
                                          utils.wait_user_input_linux):
                    startup_waited = int(False or is_cmd_success)
                    break

                if not is_cmd_success:
                    if n_retries is not None:
                        if retries == n_retries:
                            break
                        retries += 1

                    time.sleep(retry_interval)
            except KeyboardInterrupt:
                # Disable the CTRL+C interrupt while waiting for startup, otherwise terminal will close.
                pass
            except APIError:
                return 2

        return startup_waited

    @staticmethod
    def copy_files(machine_api_object: docker.models.containers.Container, path: str, tar_data: bytes) -> None:
        """Copy the files contained in tar_data in the Docker container path specified by the machine_api_object.

        Args:
            machine_api_object (docker.models.containers.Container): A Docker container.
            path (str): The path of the container where copy the tar_data.
            tar_data (bytes): The data to copy in the container.

        Returns:
            None
        """
        machine_api_object.put_archive(path, tar_data)

    @staticmethod
    def retrieve_files(machine_api_object: docker.models.containers.Container, src: str, dst: str) -> None:
        """Get the file or directory from the Docker container specified by the machine_api_object into dst.

        Args:
            machine_api_object (docker.models.containers.Container): A Docker container.
            src (str): The path of the file or folder to copy from the device.
            dst (str): The destination path on the host.

        Returns:
            None
        """
        # Get the tar from the container
        bits, stats = machine_api_object.get_archive(src)

        # Create a tmp tar file and write the chunks from the container
        with tempfile.NamedTemporaryFile(mode='wb+', suffix='.tar') as temp_file:
            for chunk in bits:
                temp_file.write(chunk)

            # After writing, go at the beginning of file
            temp_file.seek(0)

            # Now, we need to extract the tmp tar file into the destination folder
            with tarfile.open(fileobj=temp_file, mode='r') as tar_file:
                tar_file.extractall(path=dst)

    @privileged
    def get_machines_api_objects_by_filters(self, lab_hash: str = None, machine_name: str = None, user: str = None) -> \
            List[docker.models.containers.Container]:
        """Return the Docker containers objects specified by lab_hash and user.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the devices in the scenario.
            machine_name (str): The name of a device. If specified, return the specified container of the scenario.
            user (str): The name of a user on the host. If specified, return only the containers of the user.

        Returns:
            List[docker.models.containers.Container]: A list of Docker containers objects.
        """
        filters = {"label": ["app=kathara"]}
        if user:
            filters["label"].append(f"user={user}")
        if lab_hash:
            filters["label"].append(f"lab_hash={lab_hash}")
        if machine_name:
            filters["label"].append(f"name={machine_name}")

        return self.client.containers.list(all=True, filters=filters, ignore_removed=True)

    def get_machines_stats(self, lab_hash: str = None, machine_name: str = None, user: str = None) -> \
            Generator[Dict[str, DockerMachineStats], None, None]:
        """Return a generator containing the Docker devices' stats.

        Args:
            lab_hash (str): The hash of a network scenario. If specified, return all the stats of the devices in the
                scenario.
            machine_name (str): The name of a device. If specified, return the specified device stats.
            user (str): The name of a user on the host. If specified, return only the stats of the specified user.

        Returns:
            Generator[Dict[str, DockerMachineStats], None, None]: A generator containing device names as keys and
            DockerMachineStats as values.

        Raises:
            PrivilegeError: If user param is None and the user does not have root privileges.
        """
        if user is None and not utils.is_admin():
            raise PrivilegeError("You must be root to get devices statistics of all users.")

        machines_stats = {}

        @privileged
        def load_machine_stats(machine):
            if machine.name not in machines_stats:
                machines_stats[machine.name] = DockerMachineStats(machine)

        while True:
            containers = self.get_machines_api_objects_by_filters(
                lab_hash=lab_hash, machine_name=machine_name, user=user
            )
            if not containers:
                yield dict()

            pool_size = utils.get_pool_size()
            items = utils.chunk_list(containers, pool_size)
            with Pool(pool_size) as machines_pool:
                for chunk in items:
                    machines_pool.map(func=load_machine_stats, iterable=chunk)

            machines_to_remove = []
            for machine_id, machine_stats in machines_stats.items():
                try:
                    machine_stats.update()
                except StopIteration:
                    machines_to_remove.append(machine_id)
                    continue

            for k in machines_to_remove:
                machines_stats.pop(k, None)

            yield machines_stats

    @staticmethod
    def get_container_name(name: str, lab_hash: str) -> str:
        """Return the name of a Docker container.

        Args:
            name (str): The name of a Kathara device.
            lab_hash (str): The hash of a running scenario.

        Returns:
            str: The name of the Docker container in the format "|dev_prefix|_|username_prefix|_|name|_|lab_hash|".
        """
        lab_hash = lab_hash if "_%s" % lab_hash else ""
        return "%s_%s_%s_%s" % (Setting.get_instance().device_prefix, utils.get_current_user_name(), name, lab_hash)

    def _delete_machine(self, container: docker.models.containers.Container) -> None:
        """Remove a running Docker container.

        Args:
            container (docker.models.containers.Container): The Docker container to remove.

        Returns:
            None
        """
        # Build the shutdown command string
        shutdown_commands_string = "; ".join(SHUTDOWN_COMMANDS).format(machine_name=container.labels["name"])

        logging.debug(f"Executing shutdown commands on `{container.labels['name']}`: {shutdown_commands_string}")
        # Execute the shutdown commands inside the container (only if it's running)
        if container.status == "running":
            try:
                self._exec_run(container,
                               cmd=[container.labels['shell'], '-c', shutdown_commands_string],
                               stdout=True,
                               stderr=False,
                               privileged=True,
                               detach=False
                               )
            except MachineBinaryError as e:
                logging.warning(f"Shell `{e.binary}` not found in "
                                f"image `{container.image.tags[0]}` of device `{container.labels['name']}`. "
                                f"Shutdown commands will not be executed."
                                )

        container.remove(v=True, force=True)



import logging
from typing import Union, List, Set

import docker.models.images
from docker import DockerClient
from docker.errors import APIError

from ... import utils
from ...event.EventDispatcher import EventDispatcher
from ...exceptions import InvalidImageArchitectureError, DockerImageNotFoundError


class DockerImage(object):
    """Class responsible for interacting with Docker Images."""
    __slots__ = ['client']

    def __init__(self, client: DockerClient) -> None:
        self.client: DockerClient = client

    def get_local(self, image_name: str) -> docker.models.images.Image:
        """Return the specified Docker Image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            docker.models.images.Image: A Docker Image
        """
        return self.client.images.get(image_name)

    def get_remote(self, image_name: str) -> docker.models.images.RegistryData:
        """Gets the registry data for an image.

        Args:
            image_name (str): The name of the image.

        Returns:
            docker.models.images.RegistryData: The data object.

        Raises:
            `docker.errors.APIError`: If the server returns an error.
        """
        return self.client.images.get_registry_data(image_name)

    def pull(self, image_name: str) -> None:
        """Pull the specified Docker Image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            None
        """
        # If no tag or sha key is specified, we add "latest"
        if (':' or '@') not in image_name:
            image_name = "%s:latest" % image_name

        EventDispatcher.get_instance().dispatch("docker_pull_started")
        logging.info("Pulling image `%s`... This may take a while." % image_name)
        response = self.client.api.pull(image_name, stream=True, decode=True)
        for progress in response:
            EventDispatcher.get_instance().dispatch("docker_pull_progress", progress=progress)
        EventDispatcher.get_instance().dispatch("docker_pull_ended")

    def check_for_updates(self, image_name: str) -> None:
        """Update the specified image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            None
        """
        logging.debug(f"Checking updates for {image_name}...")

        if '@' in image_name:
            logging.debug(f"No need to check image digest of {image_name}.")
            return

        local_image_info = self.get_local(image_name)
        # Image has been built locally, so there's nothing to compare.
        local_repo_digests = local_image_info.attrs["RepoDigests"]
        if not local_repo_digests:
            logging.debug(f"Image {image_name} is built locally.")
            return

        remote_image_info = self.get_remote(image_name).attrs['Descriptor']
        local_repo_digest = local_repo_digests[0]
        remote_image_digest = remote_image_info["digest"]

        # Format is image_name@sha256, so we strip the first part.
        (_, local_image_digest) = local_repo_digest.split("@")
        # We only need to update tagged images, not the ones with digests.
        if remote_image_digest != local_image_digest:
            EventDispatcher.get_instance().dispatch("docker_image_update_found",
                                                    docker_image=self,
                                                    image_name=image_name)

    def check(self, image_name: str) -> None:
        """Check the existence of the specified image.

        Args:
            image_name (str): The name of a Docker Image.

        Returns:
            None

        Raises:
            ConnectionError: If there is a connection error while pulling the Docker image from Docker Hub.
            DockerImageNotFoundError: If the Docker image is not available neither on Docker Hub nor in local repository.
        """
        self._check_and_pull(image_name, pull=False)

    def check_from_list(self, images: Union[List[str], Set[str]]) -> None:
        """Check a list of specified images.

        Args:
            images (Union[List[str], Set[str]]): A list of Docker images name to pull.

        Returns:
            None
        """
        for image in images:
            self._check_and_pull(image)

    def _check_and_pull(self, image_name: str, pull: bool = True) -> None:
        """Check and pull of the specified image.

        Args:
            image_name (str): The name of a Docker Image.
            pull (bool): If True, pull the image from Docker Hub.

        Returns:
            None

        Raises:
            ConnectionError: If there is a connection error while pulling the Docker image from Docker Hub.
            DockerImageNotFoundError: If the Docker image is not available neither on Docker Hub
                nor in local repository.
            InvalidImageArchitectureError: If the Docker image is not compatible with the host architecture.
        """
        try:
            # Tries to get the image from the local Docker repository.
            image = self.get_local(image_name)
            self._check_image_architecture(image_name, image)
            try:
                if pull:
                    self.check_for_updates(image_name)
            except APIError:
                logging.debug("Cannot check updates, skipping...")
        except InvalidImageArchitectureError as e:
            raise e
        except APIError:
            # If not found, tries on Docker Hub.
            try:
                # If the image exists on Docker Hub, pulls it.
                registry_data = self.get_remote(image_name)
                self._check_image_architecture(image_name, registry_data)
                if pull:
                    self.pull(image_name)
            except APIError as e:
                if e.response.status_code == 500 and 'dial tcp' in e.explanation:
                    raise ConnectionError(
                        f"Docker Image `{image_name}` is not available in local repository and "
                        "no Internet connection is available to pull it from Docker Hub."
                    )
                else:
                    raise DockerImageNotFoundError(image_name)
            except InvalidImageArchitectureError as e:
                raise e

    @staticmethod
    def _check_image_architecture(image_name: str,
                                  image: Union[docker.models.images.Image, docker.models.images.RegistryData]) -> None:
        """Check if the specified image is compatible with the host architecture.

        Args:
            image_name (str): The name of the Docker Image to check.
            image (Union[docker.models.images.Image, docker.models.images.RegistryData]): Docker Image object.

        Returns:
            None
            
        Raises: 
            InvalidImageArchitectureError: If the Docker image is not compatible with the architecture.
        """
        host_arch = utils.get_architecture()

        # amd64 images are compatible on macOS using Rosetta.
        compatible_archs = utils.exec_by_platform(
            lambda: {host_arch}, lambda: {host_arch}, lambda: {host_arch, "amd64"}
        )

        logging.debug(f"Platform compatible architectures: {compatible_archs}")

        is_compatible = False
        if isinstance(image, docker.models.images.Image):
            is_compatible = image.attrs['Architecture'] in compatible_archs
        elif isinstance(image, docker.models.images.RegistryData):
            image_archs = list(
                filter(lambda x: x['architecture'] in compatible_archs, image.attrs['Platforms'])
            )
            is_compatible = len(image_archs) > 0
            logging.debug(f"Found compatible architectures: {image_archs}")

        if not is_compatible:
            raise InvalidImageArchitectureError(image_name, host_arch)



import logging
import os
from abc import ABC, abstractmethod
from typing import List, Dict, Any

from rich.console import Console
from rich.theme import Theme

from ....setting.Setting import Setting, SETTINGS_FILENAME
from ..CliArgs import CliArgs


class Command(ABC):
    __slots__ = ['parser', 'console']

    def __init__(self) -> None:
        self.console: Console = Console(
            theme=Theme({
                "kathara.lab_name": "bold dark_orange3",
                "kathara.lab_description": "bold dark_orange3",
                "kathara.lab_version": "bold dark_orange3",
                "kathara.lab_author": "bold dark_orange3",
                "kathara.lab_email": "bold dark_orange3",
                "kathara.lab_web": "bold dark_orange3",
            })
        )

    @abstractmethod
    def run(self, current_path: str, argv: List[str]) -> int:
        raise NotImplementedError("You must implement `run` method.")

    def parse_args(self, argv: List[str]) -> None:
        args = self.parser.parse_args(argv)
        CliArgs.get_instance().args = vars(args)

    @staticmethod
    def get_args() -> Dict[str, Any]:
        return CliArgs.get_instance().args

    @staticmethod
    def _load_custom_configuration(lab_path: str) -> None:
        """Load custom configuration from the given lab path if it exists.

        Args:
            lab_path (str): Path to the lab directory
        """
        custom_conf_path = os.path.join(lab_path, SETTINGS_FILENAME)
        if os.path.exists(custom_conf_path):
            logging.info(f'Loading custom kathara.conf file from path `{custom_conf_path}`...')
            Setting.get_instance().load_from_disk(lab_path)



from abc import ABC, abstractmethod
from typing import Optional

from .....cli.ui.setting import SettingsMenuFactory
from .....exceptions import InvocationError
from .....trdparty.consolemenu import ConsoleMenu, MenuFormatBuilder


class OptionsHandler(ABC):
    __slots__ = ['menu_factory']

    def __init__(self) -> None:
        self.menu_factory: Optional['SettingsMenuFactory.SettingsMenuFactory'] = None

    def add(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        if not self.menu_factory:
            raise InvocationError("`add` called without a MenuFactory set.")

        self.add_items(current_menu, menu_formatter)

    @abstractmethod
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        raise NotImplementedError("You must implement `add_items` method.")

    def set_menu_factory(self, menu_factory: 'SettingsMenuFactory.SettingsMenuFactory') -> None:
        self.menu_factory = menu_factory



"""
nsenter - run program with namespaces of other processes
Adapted from: https://github.com/zalando/python-nsenter
"""

import ctypes
import ctypes.util
import errno
import logging
import os
from contextlib import ExitStack
from pathlib import Path
from typing import Union, Optional, List

NAMESPACE_NAMES = frozenset(['mnt', 'ipc', 'net', 'pid', 'user', 'uts'])


class Namespace(object):
    """A context manager for entering namespaces

    Args:
        pid: The PID for the owner of the namespace to enter, or an absolute
             path to a file which represents a namespace handle.

        ns_type: The type of namespace to enter must be one of
                 mnt ipc net pid user uts.  If pid is an absolute path, this
                 much match the type of namespace it represents

        proc: The path to the /proc file system.  If running in a container
              the host proc file system may be binded mounted in a different
              location

    Raises:
        IOError: A non existent PID was provided
        ValueError: An improper ns_type was provided
        OSError: Unable to enter or exit the namespace

    Example:
        with Namespace(916, 'net'):
            #do something in the namespace
            pass

        with Namespace('/var/run/netns/foo', 'net'):
            #do something in the namespace
            pass
    """

    _log = logging.getLogger(__name__)
    _libc = ctypes.CDLL(ctypes.util.find_library('c'), use_errno=True)

    def __init__(self, pid, ns_type, proc='/proc'):
        if ns_type not in NAMESPACE_NAMES:
            raise ValueError('ns_type must be one of {0}'.format(
                ', '.join(NAMESPACE_NAMES)
            ))

        self.pid = pid
        self.ns_type = ns_type
        self.proc = proc

        # if it's numeric, then it's a pid, else assume it's a path
        try:
            pid = int(pid)
            self.target_fd = self._nsfd(pid, ns_type).open()
        except ValueError:
            self.target_fd = Path(pid).open()

        self.target_fileno = self.target_fd.fileno()

        self.parent_fd = self._nsfd('self', ns_type).open()
        self.parent_fileno = self.parent_fd.fileno()

    __init__.__annotations__ = {'pid': str, 'ns_type': str}

    def _nsfd(self, pid, ns_type):
        """Utility method to build a pathlib.Path instance pointing at the
        requested namespace entry

        Args:
            pid: The PID
            ns_type: The namespace type to enter

        Returns:
             pathlib.Path pointing to the /proc namespace entry
        """
        return Path(self.proc) / str(pid) / 'ns' / ns_type

    _nsfd.__annotations__ = {'process': str, 'ns_type': str, 'return': Path}

    def _close_files(self):
        """Utility method to close our open file handles"""
        try:
            self.target_fd.close()
        except:
            pass

        if self.parent_fd is not None:
            self.parent_fd.close()

    def __enter__(self):
        self._log.debug('Entering %s namespace %s', self.ns_type, self.pid)

        if self._libc.setns(self.target_fileno, 0) == -1:
            e = ctypes.get_errno()
            self._close_files()
            raise OSError(e, errno.errorcode[e])

    def __exit__(self, type, value, tb):
        self._log.debug('Leaving %s namespace %s', self.ns_type, self.pid)

        if self._libc.setns(self.parent_fileno, 0) == -1:
            e = ctypes.get_errno()
            self._close_files()
            raise OSError(e, errno.errorcode[e])

        self._close_files()


def nsenter(target: Union[str, int], command: str, proc: str = "/proc",
            ns_types: Optional[List[str]] = None, all_ns_types: bool = True):
    """Enter namespace, adapted from main() function to work from Python.

    Args:
        target (Union[str, int]: The PID for the owner of the namespace to enter, or an absolute
            path to a file which represents a namespace handle.
        command (str): The command to run inside the namespace.
        proc (str): The path to the /proc file system.  If running in a container the host proc file system may be
            bind-mounted in a different location.
        ns_types (Optional[str]): The type of namespaces to enter. Must be one of:
            mnt ipc net pid user uts. If target is an absolute path, much match the type of namespace it represents.
            Can be used alternatively to all.
        all_ns_types (bool): Enable all types of namespaces. Can be used alternatively to ns_types. Default is True.

    Raises:
        ValueError: Neither ns_types or all was provided.
        OSError: Unable to enter or exit the namespace.
    """
    if ns_types is None:
        ns_types = []

    if not ns_types and not all_ns_types:
        raise ValueError('You must specify at least one namespace.')

    # Disable all_ns_types in case there is the list.
    if ns_types:
        all_ns_types = False

    try:
        with ExitStack() as stack:
            namespaces = []
            for ns in NAMESPACE_NAMES:
                if ns in ns_types or all_ns_types:
                    namespaces.append(Namespace(target, ns, proc=proc))

            for ns in namespaces:
                stack.enter_context(ns)

            # Added: chroot in the namespace root, copied from nsenter util
            wd_fd = os.open(".", os.O_RDONLY)
            host_root_fd = os.open("/", os.O_RDONLY)
            root_fd = os.open(os.path.join(proc, str(target), "root"), os.O_RDONLY)

            os.fchdir(root_fd)
            os.chroot(".")
            os.close(root_fd)

            os.system(command)

        os.fchdir(host_root_fd)
        os.chroot(".")
        os.fchdir(wd_fd)

        os.close(host_root_fd)
        os.close(wd_fd)
    except OSError as exc:
        raise OSError('Unable to enter {0} namespace: {1}.'.format(ns.ns_type, exc))



from .HandleDockerImagePull import HandleDockerImagePull
from .HandleMachineTerminal import HandleMachineTerminal
from .HandleProgressBar import HandleProgressBar
from .MountDevicesVolumes import MountDevicesVolumes
from .UpdateDockerImage import UpdateDockerImage
from ....event.EventDispatcher import EventDispatcher


def register_cli_events() -> None:
    """Register events to handle UI from CLI.

    Returns:
        None
    """
    _register_link_events()
    _register_machine_events()
    _register_docker_image_pull_events()

    EventDispatcher.get_instance().register("docker_image_update_found", UpdateDockerImage())


def unregister_cli_events() -> None:
    """Unregister events from handle UI from CLI.

    Returns.
        None
"""
    EventDispatcher.get_instance().unregister("links_deploy_started")
    EventDispatcher.get_instance().unregister("link_deployed")
    EventDispatcher.get_instance().unregister("links_deploy_ended")

    EventDispatcher.get_instance().unregister("links_undeploy_started")
    EventDispatcher.get_instance().unregister("link_undeployed")
    EventDispatcher.get_instance().unregister("links_undeploy_ended")

    EventDispatcher.get_instance().unregister("machines_deploy_started")
    EventDispatcher.get_instance().unregister("machine_deployed")
    EventDispatcher.get_instance().unregister("machines_deploy_ended")

    EventDispatcher.get_instance().unregister("machines_undeploy_started")
    EventDispatcher.get_instance().unregister("machine_undeployed")
    EventDispatcher.get_instance().unregister("machines_undeploy_ended")

    EventDispatcher.get_instance().unregister("machine_deployed")
    EventDispatcher.get_instance().unregister("machine_startup_wait_started")
    EventDispatcher.get_instance().unregister("machine_startup_wait_ended")

    EventDispatcher.get_instance().unregister("machines_with_volumes")

    EventDispatcher.get_instance().unregister("docker_pull_started")
    EventDispatcher.get_instance().unregister("docker_pull_progress")
    EventDispatcher.get_instance().unregister("docker_pull_ended")

    EventDispatcher.get_instance().unregister("docker_image_update_found")


def _register_link_events() -> None:
    link_deploy_progress_bar_handler = HandleProgressBar('Deploying collision domains')
    EventDispatcher.get_instance().register("links_deploy_started", link_deploy_progress_bar_handler, "init")
    EventDispatcher.get_instance().register("link_deployed", link_deploy_progress_bar_handler, "update")
    EventDispatcher.get_instance().register("links_deploy_ended", link_deploy_progress_bar_handler, "finish")

    link_undeploy_progress_bar_handler = HandleProgressBar('Deleting collision domains')
    EventDispatcher.get_instance().register("links_undeploy_started", link_undeploy_progress_bar_handler, "init")
    EventDispatcher.get_instance().register("link_undeployed", link_undeploy_progress_bar_handler, "update")
    EventDispatcher.get_instance().register("links_undeploy_ended", link_undeploy_progress_bar_handler, "finish")


def _register_machine_events() -> None:
    machine_deploy_progress_bar_handler = HandleProgressBar('Deploying devices')
    EventDispatcher.get_instance().register("machines_deploy_started", machine_deploy_progress_bar_handler, "init")
    EventDispatcher.get_instance().register("machine_deployed", machine_deploy_progress_bar_handler, "update")
    EventDispatcher.get_instance().register("machines_deploy_ended", machine_deploy_progress_bar_handler, "finish")

    machine_undeploy_progress_bar_handler = HandleProgressBar('Deleting devices')
    EventDispatcher.get_instance().register("machines_undeploy_started", machine_undeploy_progress_bar_handler, "init")
    EventDispatcher.get_instance().register("machine_undeployed", machine_undeploy_progress_bar_handler, "update")
    EventDispatcher.get_instance().register("machines_undeploy_ended", machine_undeploy_progress_bar_handler, "finish")

    machine_terminal_handler = HandleMachineTerminal()
    EventDispatcher.get_instance().register("machine_deployed", machine_terminal_handler, "run")
    EventDispatcher.get_instance().register("machine_startup_wait_started", machine_terminal_handler, "print_wait_msg")
    EventDispatcher.get_instance().register("machine_startup_wait_ended", machine_terminal_handler, "flush")

    EventDispatcher.get_instance().register("machines_with_volumes", MountDevicesVolumes())


def _register_docker_image_pull_events():
    docker_pull_handler = HandleDockerImagePull()
    EventDispatcher.get_instance().register("docker_pull_started", docker_pull_handler, "init")
    EventDispatcher.get_instance().register("docker_pull_progress", docker_pull_handler, "update")
    EventDispatcher.get_instance().register("docker_pull_ended", docker_pull_handler, "finish")



import os
from glob import glob

from ...model.Lab import Lab
from ...utils import RESERVED_MACHINE_NAMES


class FolderParser(object):
    """Class responsible for parsing the folder in the network scenario directory."""

    @staticmethod
    def parse(path: str) -> Lab:
        """Parse the network scenario folders and return a network scenario containing the corresponding devices.

        Args:
            path (str): The path to the network scenario directory.

        Returns:
            Kathara.model.Lab.Lab: A Kathara network scenario.
        """
        lab = Lab(None, path=path)

        # Get all subfolders of lab path
        machine_folders = glob("%s/*/" % path)

        for machine_folder in machine_folders:
            # Get tail of the path, :-1 is required to remove the trailing slash
            # Otherwise basename will return an empty string
            machine_name = os.path.basename(machine_folder[:-1])

            # Shared is a reserved name, ignore it.
            if machine_name in RESERVED_MACHINE_NAMES:
                continue

            lab.get_or_new_machine(machine_name)

        return lab



from typing import List, Dict, Any


class OptionParser(object):
    """Class responsible for parsing the network scenario options from CLI."""

    @staticmethod
    def parse(options: List[str]) -> Dict[str, Any]:
        """Parse the CLI options and return a Dict containing their values.
        
        Args:
            options (List[str]): A list of string taken from CLI.

        Returns:
            Dict[str, Any]: Keys are option name and values are option value.
        """
        try:
            parsed_options = {}

            if not options:
                return parsed_options

            for option in options:
                (key, value) = option.replace('"', '').replace("'", '').split("=")
                parsed_options[key] = value

            return parsed_options
        except Exception as e:
            raise ValueError("Option parameter not valid: %s." % str(e))



from typing import Any

from .session.KubernetesWSTerminalSession import KubernetesWSTerminalSession
from ....foundation.manager.terminal.core.TerminalRunner import TerminalRunner
from ....utils import exec_by_platform


class KubernetesWSTerminal(object):
    """High-level terminal runner for Kubernetes over WebSockets.

    Args:
        handler (Any): The exec session handler.
    """

    __slots__ = ["_runner"]

    def __init__(self, handler: Any) -> None:
        session = KubernetesWSTerminalSession(handler)

        def unix():
            from ....foundation.manager.terminal.console.UnixConsoleAdapter import UnixConsoleAdapter
            return UnixConsoleAdapter()

        def windows():
            from ....foundation.manager.terminal.console.WindowsConsoleAdapter import WindowsConsoleAdapter
            return WindowsConsoleAdapter()

        console = exec_by_platform(unix, windows, unix)
        self._runner = TerminalRunner(console=console, session=session)

    def start(self) -> None:
        """Start the interactive terminal session.

        Returns:
            None
        """
        self._runner.start()



import logging
from typing import Optional, Iterable

from kubernetes import client, watch
from kubernetes.client.api import core_v1_api
from kubernetes.client.rest import ApiException

from ...model.Lab import Lab


class KubernetesNamespace(object):
    """Class responsible for interacting with Kubernetes namespaces."""

    __slots__ = ['client', 'kubernetes_secret']

    def __init__(self) -> None:
        self.client: core_v1_api.CoreV1Api = core_v1_api.CoreV1Api()

    def create(self, lab: Lab) -> Optional[client.V1Namespace]:
        """Create a Kubernetes namespace for a Kathara network scenario.

        Args:
            lab (Kathara.model.Lab.Lab): A Kathara network scenario.

        Returns:
            Optional[client.V1Namespace]: The Kubernetes namespace for the network scenario.
        """
        namespace_definition = client.V1Namespace(
            metadata=client.V1ObjectMeta(name=lab.hash, labels={'app': 'kathara'})
        )

        try:
            self.client.create_namespace(namespace_definition)
            self._wait_namespace_creation(lab.hash)

            return namespace_definition
        except ApiException:
            return None

    def undeploy(self, lab_hash: str = None) -> None:
        """Delete the Kubernetes namespace corresponding to the lab_hash.

        Args:
            lab_hash (str): The hash of a Kathara network scenario.

        Returns:
            None
        """
        try:
            self.client.delete_namespace(lab_hash)
            self._wait_namespaces_deletion(label_selector=f"kubernetes.io/metadata.name={lab_hash}")
        except ApiException:
            return

    def wipe(self) -> None:
        """Delete all the Kathara Kubernetes namespaces.

        Returns:
            None
        """
        namespaces = self.get_all()

        for namespace in namespaces:
            self.client.delete_namespace(namespace.metadata.name)

        self._wait_namespaces_deletion(label_selector="app=kathara")

    def get_all(self) -> Iterable[client.V1Namespace]:
        """Return an Iterable containing all the Kubernetes namespaces related to Kathara.

        Returns:
            Iterable[client.V1Namespace]: an Iterable containing all the Kubernetes namespaces related to Kathara.
        """
        return self.client.list_namespace(label_selector="app=kathara").items

    def get_namespace(self, lab_hash: str) -> Optional[client.V1Namespace]:
        """Return an Iterable containing all the Kubernetes namespaces related to Kathara.

        Returns:
            Iterable[client.V1Namespace]: an Iterable containing all the Kubernetes namespaces related to Kathara.
        """
        namespace = self.client.list_namespace(label_selector=f"kubernetes.io/metadata.name={lab_hash}").items
        return namespace.pop() if namespace else None

    def _wait_namespace_creation(self, lab_hash: str) -> None:
        """Wait the creation of the specified Kubernetes Namespace. Returns when the namespace becomes `Active`.

        Args:
            lab_hash (str): The name of the Kubernetes Namespace to wait.

        Returns:
            None
        """
        w = watch.Watch()
        for event in w.stream(self.client.list_namespace,
                              label_selector=f"kubernetes.io/metadata.name={lab_hash}"):
            logging.debug(f"Event: {event['type']} - Namespace: {event['object'].metadata.name}")

            if event['object'].status.phase == 'Active':
                w.stop()

    def _wait_namespaces_deletion(self, label_selector: str) -> None:
        """Wait the deletion of the specified Kubernetes Namespaces. Return when specified namespaces are terminated.

        Args:
            label_selector (str): The label used to select the namespaces.

        Returns:
            None
        """
        namespaces_to_delete = len(self.client.list_namespace(label_selector=label_selector).items)

        if namespaces_to_delete > 0:
            w = watch.Watch()
            deleted_namespaces = 0
            for event in w.stream(self.client.list_namespace, label_selector=label_selector):
                logging.debug(f"Event: {event['type']} - Namespace: {event['object'].metadata.name}")

                if event['type'] == "DELETED":
                    deleted_namespaces += 1

                if deleted_namespaces == namespaces_to_delete:
                    w.stop()



import logging
import sys


class MenuBorderStyle(object):
    """
    Base class for console menu border. Each property should be overridden by a subclass.
    """

    @property
    def bottom_left_corner(self):
        """ The outer, bottom left corner of the menu. """
        raise NotImplementedError()

    @property
    def bottom_right_corner(self):
        """ The outer, bottom right corner of the menu. """
        raise NotImplementedError()

    @property
    def inner_horizontal(self):
        """ The character for inner horizontal section lines. """
        raise NotImplementedError()

    @property
    def inner_vertical(self):
        """ The character for inner vertical section lines. """
        raise NotImplementedError()

    @property
    def intersection(self):
        """ The character for intersecting inner vertical and inner horizontal lines (a "+" shape). """
        raise NotImplementedError()

    @property
    def outer_horizontal(self):
        """ The character for outer, horizontal lines (the top and bottom lines of the menu)."""
        raise NotImplementedError()

    @property
    def outer_horizontal_inner_down(self):
        """ The character for a top horizontal line with a downward inner line (a "T" shape). """
        raise NotImplementedError()

    @property
    def outer_horizontal_inner_up(self):
        """ The character for a bottom horizontal line with an upward inner line (an inverted "T" shape). """
        raise NotImplementedError()

    @property
    def outer_vertical(self):
        """ The character for an outer vertical line of the menu (the left and right sides of the menu). """
        raise NotImplementedError()

    @property
    def outer_vertical_inner_left(self):
        """ The character for an outer vertical line, with a protruding inner line to the left. """
        raise NotImplementedError()

    @property
    def outer_vertical_inner_right(self):
        """ The character for an outer vertical line, with a protruding inner line to the right. """
        raise NotImplementedError()

    @property
    def top_left_corner(self):
        """ The top left corner of the menu. """
        raise NotImplementedError()

    @property
    def top_right_corner(self):
        """ The top right corner of the menu. """
        raise NotImplementedError()


class AsciiBorderStyle(MenuBorderStyle):
    """
    A Menu Border Style using only ASCII characters.
    """

    @property
    def bottom_left_corner(self): return '+'

    @property
    def bottom_right_corner(self): return '+'

    @property
    def inner_horizontal(self): return '-'

    @property
    def inner_vertical(self): return '|'

    @property
    def intersection(self): return '+'

    @property
    def outer_horizontal(self): return '-'

    @property
    def outer_horizontal_inner_down(self): return '+'

    @property
    def outer_horizontal_inner_up(self): return '+'

    @property
    def outer_vertical(self): return '|'

    @property
    def outer_vertical_inner_left(self): return '|'

    @property
    def outer_vertical_inner_right(self): return '|'

    @property
    def top_left_corner(self): return '+'

    @property
    def top_right_corner(self): return '+'


class LightBorderStyle(MenuBorderStyle):
    """
    MenuBorderStyle class using Unicode "light" box drawing characters.
    """

    @property
    def bottom_left_corner(self): return u'\u2514'

    @property
    def bottom_right_corner(self): return u'\u2518'

    @property
    def inner_horizontal(self): return u'\u2500'

    @property
    def inner_vertical(self): return u'\u2502'

    @property
    def intersection(self): return u'\u253C'

    @property
    def outer_horizontal(self): return u'\u2500'

    @property
    def outer_horizontal_inner_down(self): return u'\u252C'

    @property
    def outer_horizontal_inner_up(self): return u'\u2534'

    @property
    def outer_vertical(self): return u'\u2502'

    @property
    def outer_vertical_inner_left(self): return u'\u2524'

    @property
    def outer_vertical_inner_right(self): return u'\u251C'

    @property
    def top_left_corner(self): return u'\u250C'

    @property
    def top_right_corner(self): return u'\u2510'


class HeavyBorderStyle(MenuBorderStyle):
    """
    MenuBorderStyle class using Unicode "heavy" box drawing characters.
    """

    @property
    def bottom_left_corner(self): return u'\u2517'

    @property
    def bottom_right_corner(self): return u'\u251B'

    @property
    def inner_horizontal(self): return u'\u2501'

    @property
    def inner_vertical(self): return u'\u2503'

    @property
    def intersection(self): return u'\u254B'

    @property
    def outer_horizontal(self): return u'\u2501'

    @property
    def outer_horizontal_inner_down(self): return u'\u2533'

    @property
    def outer_horizontal_inner_up(self): return u'\u253B'

    @property
    def outer_vertical(self): return u'\u2503'

    @property
    def outer_vertical_inner_left(self): return u'\u252B'

    @property
    def outer_vertical_inner_right(self): return u'\u2523'

    @property
    def top_left_corner(self): return u'\u250F'

    @property
    def top_right_corner(self): return u'\u2513'


class HeavyOuterLightInnerBorderStyle(HeavyBorderStyle):
    """
    MenuBorderStyle class using Unicode "heavy" box drawing characters for the outer borders, and
    "light" box drawing characters for the inner borders.
    """

    @property
    def inner_horizontal(self): return u'\u2500'

    @property
    def inner_vertical(self): return u'\u2502'

    @property
    def intersection(self): return u'\u253C'

    @property
    def outer_horizontal_inner_down(self): return u'\u252F'

    @property
    def outer_horizontal_inner_up(self): return u'\u2537'

    @property
    def outer_vertical_inner_left(self): return u'\u2528'

    @property
    def outer_vertical_inner_right(self): return u'\u2520'


class DoubleLineBorderStyle(MenuBorderStyle):
    """
    MenuBorderStyle class using "double-line" box drawing characters.
    """

    @property
    def bottom_left_corner(self): return u'\u255A'

    @property
    def bottom_right_corner(self): return u'\u255D'

    @property
    def inner_horizontal(self): return u'\u2550'

    @property
    def inner_vertical(self): return u'\u2551'

    @property
    def intersection(self): return u'\u256C'

    @property
    def outer_horizontal(self): return u'\u2550'

    @property
    def outer_horizontal_inner_down(self): return u'\u2566'

    @property
    def outer_horizontal_inner_up(self): return u'\u2569'

    @property
    def outer_vertical(self): return u'\u2551'

    @property
    def outer_vertical_inner_left(self): return u'\u2563'

    @property
    def outer_vertical_inner_right(self): return u'\u2560'

    @property
    def top_left_corner(self): return u'\u2554'

    @property
    def top_right_corner(self): return u'\u2557'


class DoubleLineOuterLightInnerBorderStyle(DoubleLineBorderStyle):
    """
    MenuBorderStyle class using Unicode "double-line" box drawing characters for the outer borders, and
    "light" box drawing characters for the inner borders.
    """

    @property
    def inner_horizontal(self): return u'\u2500'

    @property
    def inner_vertical(self): return u'\u2502'

    @property
    def intersection(self): return u'\u253C'

    @property
    def outer_horizontal_inner_down(self): return u'\u2564'

    @property
    def outer_horizontal_inner_up(self): return u'\u2567'

    @property
    def outer_vertical_inner_left(self): return u'\u2562'

    @property
    def outer_vertical_inner_right(self): return u'\u255F'


class MenuBorderStyleType(object):
    """
    Defines the various menu border styles, as expected by the border factory.
    """

    ASCII_BORDER = 0
    """ int: Menu Border using pure ASCII characters. Usable on all platforms. """

    LIGHT_BORDER = 1
    """ int: Menu Border using the "light" box drawing characters. Should be usable on all platforms. """

    HEAVY_BORDER = 2
    """ int: Menu Border using the "heavy" box drawing characters.
        NOTE: On Windows, this border style will work ONLY on Python 3.6 and later.  It will raise a UnicodeEncodeError
        exception on earlier Python versions. If requesting this border style via the MenuBorderStyleFactory when on
        Windows/Python 3.5 or earlier, this border style will be substituted by the `DOUBLE_LINE_BORDER`. """

    DOUBLE_LINE_BORDER = 3
    """ int: Menu Border using "double-line" box drawing characters. """

    HEAVY_OUTER_LIGHT_INNER_BORDER = 4
    """ int: Menu Border using the "heavy" box drawing characters for the outer border elements, and "light" box-drawing
        characters for the inner border elements.
        NOTE: On Windows, this border style will work ONLY on Python 3.6 and later.  It will raise a UnicodeEncodeError
        exception on earlier Python versions. If requesting this border style via the MenuBorderStyleFactory when
        on Windows/Python 3.5 or earlier, this border style will be substituted by the `DOUBLE_LINE_BORDER`. """

    DOUBLE_LINE_OUTER_LIGHT_INNER_BORDER = 5
    """ int: Menu Border using the "double-line" box drawing characters for the outer border elements, and "light"
        box-drawing characters for the inner border elements."""


class MenuBorderStyleFactory(object):
    """
    Factory class for creating  MenuBorderStyle instances.
    """

    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)

    def create_border(self, border_style_type):
        """
        Create a new MenuBorderStyle instance based on the given border style type.

        Args:
            border_style_type (int):  an integer value from :obj:`MenuBorderStyleType`.

        Returns:
            :obj:`MenuBorderStyle`: a new MenuBorderStyle instance of the specified style.

        """
        if border_style_type == MenuBorderStyleType.ASCII_BORDER:
            return self.create_ascii_border()
        elif border_style_type == MenuBorderStyleType.LIGHT_BORDER:
            return self.create_light_border()
        elif border_style_type == MenuBorderStyleType.HEAVY_BORDER:
            return self.create_heavy_border()
        elif border_style_type == MenuBorderStyleType.DOUBLE_LINE_BORDER:
            return self.create_doubleline_border()
        elif border_style_type == MenuBorderStyleType.HEAVY_OUTER_LIGHT_INNER_BORDER:
            return self.create_heavy_outer_light_inner_border()
        elif border_style_type == MenuBorderStyleType.DOUBLE_LINE_OUTER_LIGHT_INNER_BORDER:
            return self.create_doubleline_outer_light_inner_border()
        else:
            # Use ASCII if we don't recognize the type
            self.logger.info('Unrecognized border style type: {}. Defaulting to ASCII.'.format(border_style_type))
            return self.create_ascii_border()

    def create_ascii_border(self):
        """
        Create an ASCII border style.

        Returns:
            :obj:`AsciiBorderStyle`:  a new instance of AsciiBorderStyle.
        """
        return AsciiBorderStyle()

    def create_light_border(self):
        """
        Create a border style using "light" box drawing characters.

        Returns:
            :obj:`LightBorderStyle`: a new instance of LightBorderStyle
        """
        return LightBorderStyle()

    def create_heavy_border(self):
        """
        Create a border style using "heavy" box drawing characters.

        NOTE: The Heavy border style will work on Windows ONLY when using Python 3.6 or later. If on Windows and
        using an earlier version of Python, the heavy border will be substituted with the DOUBLE_LINE_BORDER.

        Returns:
            :obj:`HeavyBorderStyle` or :obj:`DoubleLineBorderStyle`: a new instance of HeavyBorderStyle, unless on
            Windows and pre-Python 3.5, in which case a new instance of DoubleLineBorderStyle will be returned.
        """
        # Special case for Windows...
        if self.is_win_python35_or_earlier():
            return DoubleLineBorderStyle()
        # All other platforms...
        return HeavyBorderStyle()

    def create_heavy_outer_light_inner_border(self):
        """
        Create a border style using "heavy" box drawing characters for outer border elements, and "light"
        box drawing characters for inner border elements.

        NOTE: The Heavy border style will work on Windows ONLY when using Python 3.6 or later. If on Windows and
        using an earlier version of Python, the heavy border will be substituted with the DOUBLE_LINE_BORDER.

        Returns:
            :obj:`HeavyOuterLightInnerBorderStyle` or :obj:`DoubleLineOuterLightInnerBorderStyle`: a new instance of
            HeavyOuterLightInnerBorderStyle, unless on Windows and pre-Python 3.5, in which case a new instance of
            DoubleLineOuterLightInnerBorderStyle will be returned.
        """
        # Special case for Windows...
        if self.is_win_python35_or_earlier():
            return DoubleLineOuterLightInnerBorderStyle()
        # All other platforms...
        return HeavyOuterLightInnerBorderStyle()

    def create_doubleline_border(self):
        """
        Create a border style using "double-line" box drawing characters.

        Returns:
            :obj:`DoubleLineBorderStyle`: a new instance of DoubleLineBorderStyle.
        """
        return DoubleLineBorderStyle()

    def create_doubleline_outer_light_inner_border(self):
        """
        Create a border style using "double-line" box drawing characters for outer border elements, and "light"
        box drawing characters for inner border elements.

        Returns:
            :obj:`DoubleLineOuterLightInnerBorderStyle`: a new instance of DoubleLineOuterLightInnerBorderStyle
        """
        return DoubleLineOuterLightInnerBorderStyle()

    @staticmethod
    def is_win_python35_or_earlier():
        """
        Convenience method to determine if the current platform is Windows and Python version 3.5 or earlier.

        Returns:
            bool: True if the current platform is Windows and the Python interpreter is 3.5 or earlier; False otherwise.

        """
        return sys.platform.startswith("win") and sys.version_info.major < 3 or (
                    sys.version_info.major == 3 and sys.version_info.minor < 6)



import logging
import random

import libtmux


class TMUX(object):
    __slots__ = ['_sessions', '_server']

    __instance = None

    @staticmethod
    def get_instance():
        if TMUX.__instance is None:
            TMUX()

        return TMUX.__instance

    def __init__(self):
        if TMUX.__instance is not None:
            raise Exception("This class is a singleton!")
        else:
            self._sessions = {}
            self._server = libtmux.Server()

            TMUX.__instance = self

    def add_window(self, session_name, machine_name, shell, cwd=None):
        session_name = "Kathara" if not session_name else session_name

        initial_window_name = "%008x" % random.getrandbits(32)
        session, added = self.add_session(session_name, initial_window_name)

        machine_window = session.windows.get(window_name=machine_name, default=None)
        if not machine_window:
            logging.debug("Starting TMUX window for `%s`..." % machine_name)
            session.new_window(window_name=machine_name, window_shell=shell, start_directory=cwd)

        if added:
            self.kill_window(session, initial_window_name)

    def add_session(self, session_name, window_name):
        try:
            self._server.new_session(session_name, window_name=window_name)
        except libtmux.server.exc.TmuxSessionExists:
            session = self._get_session_from_server(session_name)
            self.add_session_by_name(session_name, session)
            return session, False
        except libtmux.server.exc.LibTmuxException as e:
            if 'duplicate session' in e.args[0][0]:
                session = self._get_session_from_server(session_name)
                self.add_session_by_name(session_name, session)
                return session, False

        logging.debug("Initialized TMUX session %s" % session_name)

        session = self._server.sessions.get(session_name=session_name)
        self.add_session_by_name(session_name, session)

        return session, True

    def add_session_by_name(self, session_name, session):
        if session_name not in self._sessions:
            self._sessions[session_name] = session

    def get_session_by_name(self, session_name):
        if session_name in self._sessions:
            return self._sessions[session_name]

        return None

    def _get_session_from_server(self, session_name):
        if self._server.has_session(session_name):
            return self._server.sessions.get(session_name=session_name)

        return None

    @staticmethod
    def kill_window(session, window_name):
        window = session.windows.get(window_name=window_name, default=None)

        if window:
            window.kill_window()



from kubernetes import config, client

from ...setting.Setting import Setting


class KubernetesConfig(object):
    """Class responsible for loading Kubernetes configurations."""
    @staticmethod
    def get_cluster_user() -> str:
        """Return the name of the current cluster user.

        Returns:
            str: The name of the current cluster user.
        """
        try:
            # Remote configuration is present, use the API Token as user
            configuration = client.Configuration.get_default_copy()
            return configuration.api_key['authorization']
        except KeyError:
            # In-Cluster configuration, take the context name as user
            _, current_context = config.kube_config.list_kube_config_contexts()
            return current_context['name']

    @staticmethod
    def load_kube_config() -> None:
        """Load the in-cluster config if launched on a master. Otherwise, setup the config for a remote master.

        Returns:
            None

        Raises:
            ConnectionError: If there is an error while reading the Kubernetes configuration.
        """
        try:
            config.load_kube_config()  # Try to load configuration if Megalos is launched on a k8s master.
        except Exception:  # Not on a k8s master, load Megalos setting to read remote cluster data.
            api_url = Setting.get_instance().api_server_url
            token = Setting.get_instance().api_token

            if not api_url or not token:
                raise ConnectionError("Cannot read Kubernetes configuration.")

            # Load the configuration and set it as default.
            configuration = client.Configuration()
            configuration.host = api_url
            configuration.api_key_prefix['authorization'] = "Bearer"
            configuration.api_key['authorization'] = token

            client.Configuration.set_default(configuration)



import logging
from multiprocessing.dummy import Pool

import requests

from ..exceptions import HTTPConnectionError
from ..utils import get_pool_size

DOCKER_HUB_KATHARA_IMAGES_URL: str = "https://hub.docker.com/v2/repositories/kathara/?page_size=-1"
DOCKER_HUB_KATHARA_TAGS_URL: str = "https://hub.docker.com/v2/repositories/{image_name}/tags/?page_size=-1&ordering"
REQUEST_TIMEOUT: int = 1

EXCLUDED_IMAGES: list[str] = ['megalos-bgp-manager', 'kathara', 'kathara-lab-checker']


class DockerHubApi(object):
    @staticmethod
    def get_images() -> filter:
        """Retrieve the list of available Kathara images on Docker Hub.

        This method retrieves the list of images from the Kathara Docker Hub account, excluding private images,
        plugins, and images specified in the EXCLUDED_IMAGES list.

        Returns:
            filter: A filtered list of currently available images that satisfy the specified conditions.

        Raises:
            HTTPConnectionError: If there is a connection error with the Docker Hub.
        """
        try:
            logging.debug("Getting Kathara images from Docker Hub...")
            response = requests.get(DOCKER_HUB_KATHARA_IMAGES_URL, timeout=(REQUEST_TIMEOUT, REQUEST_TIMEOUT))
        except requests.exceptions.Timeout as e:
            raise HTTPConnectionError(str(e))
        except requests.exceptions.ConnectionError as e:
            raise HTTPConnectionError(str(e))

        if response.status_code != 200:
            logging.debug(f"Docker Hub replied with status code {response.status_code}.")
            raise HTTPConnectionError(f"Docker Hub replied with status code {response.status_code}.")

        return filter(
            lambda x: not x['is_private'] and 'image' in x['content_types'] and x['name'] not in EXCLUDED_IMAGES,
            response.json()['results']
        )

    @staticmethod
    def get_tagged_images() -> list[str]:
        """Returns the list of available Kathara images on Docker Hub with all the active tags.

        Returns:
            list[str]: A list of strings representing the tagged Docker images in the
                format "kathara/{image_name}:{tag}".

        Raises:
            HTTPConnectionError: If there is a connection error with the Docker Hub.
        """
        images = list(DockerHubApi.get_images())
        tagged_images = []

        def get_image_tag(image):
            image_name = f"{image['namespace']}/{image['name']}"
            try:
                logging.debug(f"Getting tags for image `{image_name}` from Docker Hub...")
                response = requests.get(
                    DOCKER_HUB_KATHARA_TAGS_URL.format(image_name=image_name),
                    timeout=(REQUEST_TIMEOUT, REQUEST_TIMEOUT)
                )
            except requests.exceptions.Timeout as e:
                raise HTTPConnectionError(str(e))
            except requests.exceptions.ConnectionError as e:
                raise HTTPConnectionError(str(e))

            if response.status_code != 200:
                logging.debug(
                    f"Error while retrieving tags for image `{image_name}. "
                    f"Docker Hub replied with status code {response.status_code}."
                )
                raise HTTPConnectionError(
                    f"Error while retrieving tags for image `{image_name}`. "
                    f"Docker Hub replied with status code {response.status_code}."
                )

            tagged_images.extend(list(map(
                lambda x: f"{image_name}:{x['name']}" if x['name'] != "latest" else image_name,
                filter(lambda x: x['tag_status'] == 'active', response.json()['results'])
            )))

        pool_size = get_pool_size()
        with Pool(pool_size) as tags_pool:
            tags_pool.map(func=get_image_tag, iterable=images)

        return sorted(tagged_images)



from ..factory.Factory import Factory


class SettingsAddonFactory(Factory):
    def __init__(self) -> None:
        self.module_template: str = "Kathara.setting.addon"
        self.name_template: str = "%sSettingsAddon"



import argparse
from typing import List

from ..ui.utils import create_panel
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...strings import strings, wiki_description


class VcleanCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vclean',
            description=strings['vclean'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='The name of the device to clean.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        lab = Lab("kathara_vlab")

        self.console.print(
            create_panel(f"Stopping Device `{args['name']}`", style="blue bold", justify="center")
        )

        Kathara.get_instance().undeploy_lab(lab_name=lab.name, selected_machines={args['name']})

        return 0



import itertools

from . import ConsoleMenu
from .items import SubmenuItem


class MultiSelectMenu(ConsoleMenu):
    """
    Console menu that allows the selection of multiple menu items at a single prompt.

    Args:
        title: The menu title.
        subtitle: The menu subtitle.
        formatter: The menu formatter instance for styling the menu.
        prologue_text: The text to display in the prologue section of the menu.
        epilogue_text: The text to display in the epilogue section of the menu.
        show_exit_option (bool): Determines if the exit item should be displayed.
        exit_option_text (str): Text for the Exit menu item. Defaults to 'Exit'.
    """

    def __init__(self, title=None, subtitle=None, formatter=None,
                 prologue_text=None, epilogue_text=None, ack_item_completion=True,
                 show_exit_option=True, exit_option_text='Exit'):
        super(MultiSelectMenu, self).__init__(title, subtitle, formatter=formatter,
                                              prologue_text=prologue_text, epilogue_text=epilogue_text,
                                              show_exit_option=show_exit_option, exit_option_text=exit_option_text)
        self.ack_item_completion = ack_item_completion

    def append_item(self, item):
        """
        Add an item to the end of the menu before the exit item.

        Note that Multi-Select Menus will not allow a SubmenuItem to be added, as multi-select menus
        are expected to be used only for executing multiple actions.

        Args:
            item (:obj:`MenuItem`): The item to be added

        Raises:
            TypeError: If the specified MenuIem is a SubmenuItem.
        """
        if isinstance(item, SubmenuItem):
            raise TypeError("SubmenuItems cannot be added to a MultiSelectMenu")
        super(MultiSelectMenu, self).append_item(item)

    def process_user_input(self):
        """
        This overrides the method in ConsoleMenu to allow for comma-delimited and range inputs.

        Examples:
            All of the following inputs would have the same result:
                * 1,2,3,4
                * 1-4
                * 1-2,3-4
                * 1 - 4
                * 1, 2, 3, 4
        Raises:
            ValueError: If the input cannot be correctly parsed.
        """
        user_input = self.screen.input()

        try:
            indexes = self.__parse_range_list(user_input)
            # Subtract 1 from each number for its actual index number
            indexes[:] = [x - 1 for x in indexes if 0 < x < len(self.items) + 1]
            for index in indexes:
                self.current_option = index
                self.select()
        except Exception as e:
            return

    @staticmethod
    def __parse_range(rng):
        parts = rng.split('-')
        if 1 > len(parts) > 2:
            raise ValueError("Bad range: '%s'" % (rng,))
        parts = [int(i) for i in parts]
        start = parts[0]
        end = start if len(parts) == 1 else parts[1]
        if start > end:
            end, start = start, end
        return range(start, end + 1)

    def __parse_range_list(self, rngs):
        return sorted(set(itertools.chain(*[self.__parse_range(rng) for rng in rngs.split(',')])))



from .menu_borders import MenuBorderStyle, MenuBorderStyleFactory
from .menu_margins import MenuMargins
from .menu_padding import MenuPadding


class MenuStyle(object):
    """
    Class for specifying all menu styling, such as margins, padding, and border style.

    Args:
        margins (:obj:`MenuMargins`): The menu margin settings.
        padding (:obj:`MenuPadding`): The menu padding.
        border_style (:obj:MenuBorderStyle`): The menu border style. Takes precedence over `border_style_type` if both
            are specified.
        border_style_type (int): The border style type as defined by :obj:`MenuBorderStyleType`.
        border_style_factory (:obj:`MenuBorderStyleFactory`): The factory instance to use to create the borders.
    """

    def __init__(self, margins=None, padding=None, border_style=None, border_style_type=None,
                 border_style_factory=None):
        self.__margins = None
        self.__padding = None
        self.__border_style_factory = None
        self.__border_style = None

        if margins is None:
            margins = MenuMargins()
        self.margins = margins

        if padding is None:
            padding = MenuPadding()
        self.padding = padding

        if border_style_factory is None:
            border_style_factory = MenuBorderStyleFactory()
        self.border_style_factory = border_style_factory

        if border_style is not None:
            #  A specified border_style takes precedence
            self.border_style = border_style
        elif border_style_type is not None:
            # If we have a border style type, create border style with factory
            self.border_style = self.border_style_factory.create_border(border_style_type)
        else:
            # No border style class or type was given, so use default border style
            self.border_style = self.border_style_factory.create_light_border()

    @property
    def margins(self):
        """
        The margins instance.

        Returns:
            :obj:`MenuMargins`: The MenuMargins instance.
        """
        return self.__margins

    @margins.setter
    def margins(self, margins):
        if not isinstance(margins, MenuMargins):
            raise TypeError('margins must be of type MenuMargins')
        self.__margins = margins

    @property
    def padding(self):
        """
        The padding instance.

        Returns:
            :obj:`MenuPadding`: The MenuPadding instance.
        """
        return self.__padding

    @padding.setter
    def padding(self, padding):
        if not isinstance(padding, MenuPadding):
            raise TypeError('padding must be of type MenuPadding')
        self.__padding = padding

    @property
    def border_style(self):
        """
        The border style instance.

        Returns:
            :obj:`MenuBorderStyle`: The MenuBorderStyle instance.
        """
        return self.__border_style

    @border_style.setter
    def border_style(self, border_style):
        if not isinstance(border_style, MenuBorderStyle):
            raise TypeError('border_style must be of type MenuBorderStyle')
        self.__border_style = border_style

    @property
    def border_style_factory(self):
        """
        The border style factory instance.

        Returns:
            :obj:`MenuBorderStyleFactory`: The MenuBorderStyleFactory instance.

        """
        return self.__border_style_factory

    @border_style_factory.setter
    def border_style_factory(self, border_style_factory):
        if not isinstance(border_style_factory, MenuBorderStyleFactory):
            raise TypeError('border_style_factory must be of type MenuBorderStyleFactory')
        self.__border_style_factory = border_style_factory



import argparse
from typing import List

from ..ui.utils import create_panel, interface_cd_mac, volume
from ... import utils
from ...exceptions import PrivilegeError
from ...foundation.cli.command.Command import Command
from ...manager.Kathara import Kathara
from ...model.Lab import Lab
from ...setting.Setting import Setting
from ...strings import strings, wiki_description


class VstartCommand(Command):
    def __init__(self) -> None:
        Command.__init__(self)

        self.parser: argparse.ArgumentParser = argparse.ArgumentParser(
            prog='kathara vstart',
            description=strings['vstart'],
            epilog=wiki_description,
            add_help=False
        )

        self.parser.add_argument(
            '-h', '--help',
            action='help',
            default=argparse.SUPPRESS,
            help='Show a help message and exit.'
        )

        group = self.parser.add_mutually_exclusive_group(required=False)

        group.add_argument(
            "--noterminals",
            action="store_const",
            dest="terminals",
            const=False,
            default=None,
            help='Start the device without opening a terminal window.'
        )
        group.add_argument(
            "--terminals",
            action="store_const",
            dest="terminals",
            const=True,
            help='Start the device opening its terminal window.'
        )
        group.add_argument(
            '--num_terms',
            metavar='NUM_TERMS',
            required=False,
            help='Choose the number of terminals to open for the device.'
        )
        self.parser.add_argument(
            "--privileged",
            action="store_const",
            const=True,
            required=False,
            help='Start the device in privileged mode. MUST BE ROOT FOR THIS OPTION.'
        )
        self.parser.add_argument(
            '-n', '--name',
            metavar='DEVICE_NAME',
            required=True,
            help='Name of the device to be started.'
        )
        self.parser.add_argument(
            '--eth',
            type=interface_cd_mac,
            dest='eths',
            metavar='N:CD/MAC',
            nargs='+',
            required=False,
            help='Set a specific interface on a collision domain.'
        )
        self.parser.add_argument(
            '-e', '--exec',
            required=False,
            dest='exec_commands',
            nargs='*',
            help='Execute a specific command in the device during startup.'
        )
        self.parser.add_argument(
            '--mem',
            required=False,
            help='Limit the amount of RAM available for this device.'
        )
        self.parser.add_argument(
            '--cpus',
            required=False,
            help='Limit the amount of CPU available for this device.'
        )
        self.parser.add_argument(
            '-i', '--image',
            required=False,
            help='Run this device with a specific Docker Image.'
        )
        hosthome_group = self.parser.add_mutually_exclusive_group(required=False)
        hosthome_group.add_argument(
            '--no-hosthome', '-H',
            dest="hosthome_mount",
            action="store_const",
            const=False,
            help='Do not mount "/hosthome" directory inside the device.'
        )
        hosthome_group.add_argument(
            '--hosthome',
            dest="hosthome_mount",
            action="store_const",
            const=True,
            help='Mount "/hosthome" directory inside the device.'
        )
        self.parser.add_argument(
            '--terminal-emu',
            required=False,
            help='Set a different terminal emulator application (Unix only).'
        )
        self.parser.add_argument(
            '--print', '--dry-run',
            dest='dry_mode',
            required=False,
            action='store_true',
            help='Check if the device parameters are correct (dry run).'
        )
        self.parser.add_argument(
            '--bridged',
            required=False,
            action='store_true',
            help='Add a bridge interface to the device.'
        )
        self.parser.add_argument(
            '--port',
            dest='ports',
            metavar='[HOST:]GUEST[/PROTOCOL]',
            nargs='+',
            required=False,
            help='Map localhost port HOST to the internal port GUEST of the device for the specified PROTOCOL.'
        )
        self.parser.add_argument(
            '--sysctl',
            dest='sysctls',
            metavar='SYSCTL',
            nargs='+',
            required=False,
            help='Set sysctl option for the device.'
        )
        self.parser.add_argument(
            '--env',
            dest='envs',
            metavar='ENV',
            nargs='+',
            required=False,
            help='Set environment variable for the device.'
        )
        self.parser.add_argument(
            '--ulimit',
            dest='ulimits',
            metavar='KEY=SOFT[:HARD]',
            nargs='+',
            required=False,
            help='Set ulimit for the device.'
        )
        self.parser.add_argument(
            '--volume',
            type=volume,
            dest='volumes',
            metavar='HOST|GUEST|[MODE]',
            nargs='+',
            required=False,
            help='Specify a volume to mount.'
        )
        self.parser.add_argument(
            '--shell',
            required=False,
            help='Set the shell (sh, bash, etc.) that should be used inside the device.'
        )
        self.parser.add_argument(
            '--entrypoint',
            metavar='ENTRYPOINT',
            required=False,
            help='Specify the entrypoint command of the device.'
        )
        self.parser.add_argument(
            "args",
            metavar='ARG',
            nargs=argparse.REMAINDER,
            help='Specify extra arguments for the entrypoint command.'
        )

    def run(self, current_path: str, argv: List[str]) -> int:
        self.parse_args(argv)
        args = self.get_args()

        name = args.pop('name')

        self.console.print(
            create_panel(
                f"Checking Device `{name}`" if args['dry_mode'] else f"Starting Device `{name}`",
                style="blue bold", justify="center"
            )
        )

        if args['dry_mode']:
            self.console.print(f"[green]\u2713 [bold]{name}[/bold] configuration is correct.")
            return 0

        Setting.get_instance().open_terminals = args['terminals'] if args['terminals'] is not None \
            else Setting.get_instance().open_terminals
        Setting.get_instance().terminal = args['terminal_emu'] or Setting.get_instance().terminal
        Setting.get_instance().device_shell = args['shell'] or Setting.get_instance().device_shell

        lab = Lab("kathara_vlab")
        lab.add_option('hosthome_mount', args['hosthome_mount'])
        lab.add_option('shared_mount', False)

        if args['privileged']:
            if not utils.is_admin():
                raise PrivilegeError("You must be root in order to start this Kathara device in privileged mode.")
            else:
                if Setting.get_instance().open_terminals:
                    self.console.print(
                        "[yellow]\u26a0 Running devices with privileged capabilities, terminal might not open!"
                    )
        lab.add_global_machine_metadata('privileged', args['privileged'])

        if args['args'] and args['args'][0] == "--":
            args['args'] = args['args'][1:]

        volumes = args.pop('volumes')
        eths = args.pop('eths')

        device = lab.get_or_new_machine(name, **args)

        if eths:
            for iface_number, cd, mac_address in eths:
                try:
                    lab.connect_machine_to_link(device.name, cd,
                                                machine_iface_number=int(iface_number),
                                                mac_address=mac_address)
                except ValueError:
                    s = f"{cd}/{mac_address}" if mac_address else f"{cd}"
                    raise SyntaxError(f"Interface number in `--eth {iface_number}:{s}` is not a number.")

        if volumes:
            for v in volumes:
                device.add_meta("volume", v)

        Kathara.get_instance().deploy_lab(lab)

        return 0



from . import utils as setting_utils
from ....exceptions import HTTPConnectionError
from ....foundation.cli.ui.setting.OptionsHandler import OptionsHandler
from ....manager.Kathara import Kathara
from ....setting.Setting import Setting, DEFAULTS, AVAILABLE_DEBUG_LEVELS
from ....trdparty.consolemenu import *
from ....trdparty.consolemenu.items import *
from ....trdparty.consolemenu.validators.regex import RegexValidator
from ....utils import exec_by_platform
from ....validator.ImageValidator import ImageValidator
from ....validator.TerminalValidator import TerminalValidator
from ....webhooks.DockerHubApi import DockerHubApi

SHELLS_HINT = ["/bin/bash", "/bin/sh", "/bin/ash", "/bin/ksh", "/bin/zsh", "/bin/fish", "/bin/csh", "/bin/tcsh"]
TERMINALS_OSX = ["Terminal", "iTerm"]


class CommonOptionsHandler(OptionsHandler):
    def add_items(self, current_menu: ConsoleMenu, menu_formatter: MenuFormatBuilder) -> None:
        # Manager Submenu
        managers = Kathara.get_available_managers_name()
        manager_type = Setting.get_instance().manager_type

        choose_manager_string = "Choose default manager"
        manager_menu = SelectionMenu(
            strings=[],
            title=choose_manager_string,
            subtitle=lambda: "Current: %s" % managers[manager_type],
            prologue_text="""Manager is the Engine used to run Kathara labs.
            
                          Default is `%s`.""" % managers[DEFAULTS['manager_type']],
            formatter=menu_formatter
        )

        for name, formatted_name in managers.items():
            manager_menu.append_item(
                FunctionItem(
                    text=formatted_name,
                    function=self.update_manager_value,
                    args=[current_menu, name],
                    should_exit=True
                )
            )

        manager_item = SubmenuItem(choose_manager_string, manager_menu, current_menu)

        # Image Selection Submenu
        image_string = "Choose default image"
        select_image_menu = SelectionMenu(
            strings=[],
            title=image_string,
            subtitle=setting_utils.current_string("image"),
            prologue_text="""Default Docker image when you start a network scenario or a single Kathara device.
                          
                          Default is `%s`.""" % DEFAULTS['image'],
            formatter=menu_formatter
        )

        try:
            for image_name in DockerHubApi.get_tagged_images():
                select_image_menu.append_item(
                    FunctionItem(
                        text=image_name,
                        function=setting_utils.update_setting_value,
                        args=['image', image_name],
                        should_exit=True
                    )
                )
        except HTTPConnectionError:
            pass

        select_image_menu.append_item(
            FunctionItem(
                text="Choose another image",
                function=setting_utils.update_value,
                args=['image',
                      ImageValidator(),
                      'Write the name of a Docker image available on Docker Hub:',
                      None
                      ],
                should_exit=True
            )
        )
        submenu_item = SubmenuItem(image_string, select_image_menu, current_menu)

        # Open Terminals Option
        open_terminals_string = "Automatically open terminals on startup"
        open_terminals_menu = SelectionMenu(
            strings=[],
            title=open_terminals_string,
            subtitle=setting_utils.current_bool("open_terminals"),
            prologue_text="""Determines if the device terminal should be opened when starting it.
            
                          Default is %s."""
                          % setting_utils.format_bool(DEFAULTS['open_terminals']),
            formatter=menu_formatter
        )

        open_terminals_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=['open_terminals', True],
                should_exit=True
            )
        )
        open_terminals_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=['open_terminals', False],
                should_exit=True
            )
        )

        open_terminals_item = SubmenuItem(open_terminals_string, open_terminals_menu, current_menu)

        # Machine Shell Submenu
        machine_shell_string = "Choose device shell to be used"
        machine_shell_menu = SelectionMenu(
            strings=[],
            title=machine_shell_string,
            subtitle=setting_utils.current_string("device_shell"),
            formatter=menu_formatter,
            prologue_text="""The shell to use inside the device.
            
                          **The application must be correctly installed in the Docker image used for the device!**
                          
                          Default is `%s`, but it depends on the used Docker image.""" % DEFAULTS['device_shell']
        )

        for shell in SHELLS_HINT:
            machine_shell_menu.append_item(
                FunctionItem(
                    text=shell,
                    function=setting_utils.update_setting_value,
                    args=["device_shell", shell],
                    should_exit=True
                )
            )
        machine_shell_menu.append_item(
            FunctionItem(
                text="Choose another shell",
                function=setting_utils.update_value,
                args=['device_shell',
                      RegexValidator(r"^(\w|/)+$"),
                      'Write the name of a shell:',
                      'Shell name is not valid!'
                      ],
                should_exit=True
            )
        )

        machine_shell_item = SubmenuItem(machine_shell_string, machine_shell_menu, current_menu)

        # Terminal Emulator Submenu
        # Linux Version
        def terminal_emulator_menu_linux():
            terminal_string = "Choose terminal emulator to be used"
            terminal_menu = SelectionMenu(
                strings=[],
                title=terminal_string,
                subtitle=setting_utils.current_string("terminal"),
                formatter=menu_formatter,
                prologue_text="""Terminal emulator application to be used for device terminals.
                 
                              **The application must be correctly installed in the host system!**
                              
                              Default is `%s`.""" % DEFAULTS['terminal']
            )

            terminal_menu.append_item(
                FunctionItem(
                    text="/usr/bin/xterm",
                    function=setting_utils.update_setting_value,
                    args=["terminal", "/usr/bin/xterm"],
                    should_exit=True
                )
            )
            terminal_menu.append_item(
                FunctionItem(
                    text="/usr/bin/gnome-terminal",
                    function=setting_utils.update_setting_value,
                    args=["terminal", "/usr/bin/gnome-terminal"],
                    should_exit=True
                )
            )
            terminal_menu.append_item(
                FunctionItem(
                    text="TMUX",
                    function=setting_utils.update_setting_value,
                    args=["terminal", "TMUX"],
                    should_exit=True
                )
            )
            terminal_menu.append_item(
                FunctionItem(
                    text="Choose another terminal emulator",
                    function=setting_utils.update_value,
                    args=['terminal',
                          TerminalValidator(),
                          'Write the path of a terminal emulator:',
                          'Terminal emulator is not valid! Install it before using it.'
                          ],
                    should_exit=True
                )
            )

            return SubmenuItem(terminal_string, terminal_menu, current_menu)

        # macOS Version
        def terminal_emulator_menu_osx():
            terminal_string = "Choose terminal emulator to be used"
            terminal_menu = SelectionMenu(
                strings=[],
                title=terminal_string,
                subtitle=setting_utils.current_string("terminal"),
                formatter=menu_formatter,
                prologue_text="""Terminal emulator application to be used for device terminals.
                 
                              **The application must be correctly installed in the host system!**
                              
                              Default is `Terminal`."""
            )

            for terminal in TERMINALS_OSX:
                terminal_menu.append_item(
                    FunctionItem(
                        text=terminal,
                        function=setting_utils.update_setting_value,
                        args=["terminal", terminal],
                        should_exit=True
                    )
                )
            terminal_menu.append_item(
                FunctionItem(
                    text="TMUX",
                    function=setting_utils.update_setting_value,
                    args=["terminal", "TMUX"],
                    should_exit=True
                )
            )

            return SubmenuItem(terminal_string, terminal_menu, current_menu)

        terminal_item = exec_by_platform(terminal_emulator_menu_linux, lambda: None, terminal_emulator_menu_osx)

        # Prefixes Submenu
        prefixes_string = "Choose Kathara prefixes"
        prefixes_menu = SelectionMenu(
            strings=[],
            title=prefixes_string,
            formatter=menu_formatter,
            prologue_text="""Prefixes assigned to the network and device names when deployed.
            
                          Defaults are `%s` and `%s`.""" % (DEFAULTS['net_prefix'], DEFAULTS['device_prefix'])
        )

        net_prefix_item = FunctionItem(
            text=setting_utils.current_string("net_prefix", text="Insert Kathara networks prefix"),
            function=setting_utils.update_value,
            args=['net_prefix',
                  RegexValidator(r"^[a-z]+_?[a-z_]+$"),
                  'Write a Kathara networks prefix:',
                  'Network Prefix must only contain lowercase letters and underscore.'
                  ],
            should_exit=True
        )
        prefixes_menu.append_item(net_prefix_item)

        machine_prefix_item = FunctionItem(
            text=setting_utils.current_string("device_prefix", text="Insert Kathara devices prefix"),
            function=setting_utils.update_value,
            args=['device_prefix',
                  RegexValidator(r"^[a-z]+_?[a-z_]+$"),
                  'Write a Kathara devices prefix:',
                  'Device Prefix must only contain lowercase letters and underscore.'
                  ],
            should_exit=True
        )
        prefixes_menu.append_item(machine_prefix_item)

        prefixes_item = SubmenuItem(prefixes_string, prefixes_menu, current_menu)

        # Debug Level Submenu
        debug_level_string = "Choose logging level to be used"
        debug_level_menu = SelectionMenu(
            strings=[],
            title=debug_level_string,
            subtitle=setting_utils.current_string("debug_level"),
            formatter=menu_formatter,
            prologue_text="""Logging level of Kathara messages.
                            
                          Default is `%s`.""" % DEFAULTS['debug_level']
        )

        for debug_level in AVAILABLE_DEBUG_LEVELS:
            debug_level_menu.append_item(
                FunctionItem(
                    text=debug_level,
                    function=setting_utils.update_setting_value,
                    args=["debug_level", debug_level],
                    should_exit=True
                )
            )

        debug_level_item = SubmenuItem(debug_level_string, debug_level_menu, current_menu)

        # Print Startup Logs Option
        print_startup_log_string = "Print Startup Logs on device startup"
        print_startup_log_menu = SelectionMenu(
            strings=[],
            title=open_terminals_string,
            subtitle=setting_utils.current_bool("print_startup_log"),
            formatter=menu_formatter,
            prologue_text="""When opening a device terminal, print its startup log.
            
                          Default is %s.""" % setting_utils.format_bool(DEFAULTS['print_startup_log'])
        )

        print_startup_log_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=['print_startup_log', True],
                should_exit=True
            )
        )
        print_startup_log_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=['print_startup_log', False],
                should_exit=True
            )
        )

        print_startup_log_item = SubmenuItem(print_startup_log_string, print_startup_log_menu, current_menu)

        # Enable IPv6 Option
        enable_ipv6_string = "Enable IPv6"
        enable_ipv6_menu = SelectionMenu(
            strings=[],
            title=enable_ipv6_string,
            subtitle=setting_utils.current_bool("enable_ipv6"),
            formatter=menu_formatter,
            prologue_text="""This option enables IPv6 inside the devices.
            
                          Default is %s.""" % setting_utils.format_bool(DEFAULTS['enable_ipv6'])
        )

        enable_ipv6_menu.append_item(
            FunctionItem(
                text="Yes",
                function=setting_utils.update_setting_value,
                args=['enable_ipv6', True],
                should_exit=True
            )
        )
        enable_ipv6_menu.append_item(
            FunctionItem(
                text="No",
                function=setting_utils.update_setting_value,
                args=['enable_ipv6', False],
                should_exit=True
            )
        )

        enable_ipv6_item = SubmenuItem(enable_ipv6_string, enable_ipv6_menu, current_menu)

        # Volume Mount Policy Option
        volume_mount_policy_string = "Device Volume Mount Policy"
        volume_mount_policy_menu = SelectionMenu(
            strings=[],
            title=volume_mount_policy_string,
            subtitle=setting_utils.current_string("volume_mount_policy"),
            prologue_text="""Choose the policy to apply when a device is trying to mount a volume from the host.             

                                  \tDefault is %s.""" % DEFAULTS['volume_mount_policy'],
            formatter=menu_formatter
        )

        volume_mount_policy_menu.append_item(
            FunctionItem(
                text="Always",
                function=setting_utils.update_setting_value,
                args=["volume_mount_policy", "Always"],
                should_exit=True
            )
        )
        volume_mount_policy_menu.append_item(
            FunctionItem(
                text="Prompt",
                function=setting_utils.update_setting_value,
                args=["volume_mount_policy", "Prompt"],
                should_exit=True
            )
        )
        volume_mount_policy_menu.append_item(
            FunctionItem(
                text="Never",
                function=setting_utils.update_setting_value,
                args=["volume_mount_policy", "Never"],
                should_exit=True
            )
        )

        volume_mount_policy_item = SubmenuItem(volume_mount_policy_string, volume_mount_policy_menu, current_menu)

        current_menu.append_item(manager_item)
        current_menu.append_item(submenu_item)
        current_menu.append_item(open_terminals_item)
        current_menu.append_item(machine_shell_item)
        if terminal_item:
            current_menu.append_item(terminal_item)
        current_menu.append_item(prefixes_item)
        current_menu.append_item(debug_level_item)
        current_menu.append_item(print_startup_log_item)
        current_menu.append_item(enable_ipv6_item)
        current_menu.append_item(volume_mount_policy_item)

    def update_manager_value(self, current_menu: ConsoleMenu, value: str) -> None:
        setting_utils.update_setting_value("manager_type", value)

        # exit() does not work. So we force the menu to exit as the user selected the Exit option.
        # This is always the last item of the current menu.
        exit_option = len(current_menu.items) - 1
        current_menu.go_to(exit_option)

        clear_terminal()

        # Create a new menu (Manager could be changed so specific options need to be changed)
        new_menu = self.menu_factory.create_menu()
        new_menu.draw()
        new_menu.show()



from typing import Optional, Dict, Any

from ...foundation.setting.SettingsAddon import SettingsAddon
from ...types import SharedCollisionDomainsOption

DEFAULTS = {
    "hosthome_mount": False,
    "shared_mount": True,
    "image_update_policy": "Prompt",
    "shared_cds": SharedCollisionDomainsOption.NOT_SHARED,
    "remote_url": None,
    "cert_path": None,
    "network_plugin": "kathara/katharanp_vde"
}


class DockerSettingsAddon(SettingsAddon):
    __slots__ = ['hosthome_mount', 'shared_mount', 'image_update_policy', 'shared_cds',
                 'remote_url', 'cert_path', 'network_plugin']

    def __init__(self) -> None:
        self.hosthome_mount: bool = False
        self.shared_mount: bool = True
        self.image_update_policy: str = 'Prompt'
        self.shared_cds: int = SharedCollisionDomainsOption.NOT_SHARED
        self.remote_url: Optional[str] = None
        self.cert_path: Optional[str] = None
        self.network_plugin: Optional[str] = "kathara/katharanp_vde"

    def _to_dict(self) -> Dict[str, Any]:
        return {
            'hosthome_mount': self.hosthome_mount,
            'shared_mount': self.shared_mount,
            'image_update_policy': self.image_update_policy,
            'shared_cds': self.shared_cds,
            'remote_url': self.remote_url,
            'cert_path': self.cert_path,
            'network_plugin': self.network_plugin
        }



import asyncio
import ctypes
from ctypes import wintypes
from typing import Any, Optional, Dict, Callable

from ..core.IConsoleAdapter import IConsoleAdapter

# Low Level Win32 Handles
STD_INPUT_HANDLE: int = -10
STD_OUTPUT_HANDLE: int = -11

# Data structures for the Win32 Input Events
KEY_EVENT: int = 0x0001
WINDOW_BUFFER_SIZE_EVENT: int = 0x0004


class KEY_EVENT_RECORD(ctypes.Structure):
    """ctypes mapping of Win32 KEY_EVENT_RECORD structure."""
    _fields_ = [
        ("bKeyDown", wintypes.BOOL),
        ("wRepeatCount", wintypes.WORD),
        ("wVirtualKeyCode", wintypes.WORD),
        ("wVirtualScanCode", wintypes.WORD),
        ("uChar", wintypes.WCHAR),
        ("dwControlKeyState", wintypes.DWORD),
    ]


class COORD(ctypes.Structure):
    """ctypes mapping of Win32 COORD structure (X, Y)."""
    _fields_ = [("X", wintypes.SHORT), ("Y", wintypes.SHORT)]


class WINDOW_BUFFER_SIZE_RECORD(ctypes.Structure):
    """ctypes mapping of Win32 WINDOW_BUFFER_SIZE_RECORD structure."""
    _fields_ = [("dwSize", COORD)]


class EVENT_UNION(ctypes.Union):
    """ctypes union mapping the Win32 INPUT_RECORD.Event union."""
    _fields_ = [
        ("KeyEvent", KEY_EVENT_RECORD),
        ("WindowBufferSize", WINDOW_BUFFER_SIZE_RECORD),
    ]


class INPUT_RECORD(ctypes.Structure):
    """ctypes mapping of Win32 INPUT_RECORD structure."""
    _fields_ = [
        ("EventType", wintypes.WORD),
        ("Event", EVENT_UNION),
    ]


# Virtual Key Codes
# Taken from https://github.com/prompt-toolkit/python-prompt-toolkit/blob/main/src/prompt_toolkit/input/win32.py#L140
ANSI_ESC: str = "\x1b["
KEYCODES: Dict[int, str] = {
    # Home/End
    33: "5~",
    34: "6~",
    35: "F",
    36: "H",
    # Arrows
    37: "D",
    38: "A",
    39: "C",
    40: "B",
    45: "2~",
    46: "3~",
    # F-keys
    112: "11~",
    113: "12~",
    114: "13~",
    115: "14~",
    116: "15~",
    117: "17~",
    118: "18~",
    119: "19~",
    120: "20~",
    121: "21~",
    122: "23~",
    123: "24~",
}


class WindowsConsoleAdapter(IConsoleAdapter):
    """Windows-specific console adapter.

    Attributes:
        _term_handle (Optional[int]): Handle to STD_INPUT for ReadConsoleW.
        _orig_mode (Optional[int]): Original input mode bits to restore on exit_raw().
        _stdout_handle (int): Handle to STD_OUTPUT for WriteConsoleW.
        _input_task (Optional[asyncio.Task]): Asyncio task pumping input events.
        _resize_cb (Optional[Callable[[int, int], None]]): Callback invoked on console resize.
    """
    __slots__ = ["_term_handle", "_orig_mode", "_stdout_handle", "_input_task", "_resize_cb"]

    def __init__(self) -> None:
        self._term_handle: Optional[int] = None
        self._orig_mode: Optional[int] = None

        self._stdout_handle: int = ctypes.windll.kernel32.GetStdHandle(STD_OUTPUT_HANDLE)

        self._input_task: Optional[asyncio.Task] = None
        self._resize_cb: Optional[Callable[[int, int], None]] = None

    def enter_raw(self) -> None:
        """Put the terminal into raw mode.

        Returns:
            None
        """
        handle = ctypes.windll.kernel32.GetStdHandle(STD_INPUT_HANDLE)
        mode = wintypes.DWORD()
        if not ctypes.windll.kernel32.GetConsoleMode(handle, ctypes.byref(mode)):
            return

        old_mode = mode.value
        new_mode = old_mode
        new_mode &= ~0x0004  # ENABLE_ECHO_INPUT
        new_mode &= ~0x0002  # ENABLE_LINE_INPUT
        new_mode &= ~0x0001  # ENABLE_PROCESSED_INPUT

        if not ctypes.windll.kernel32.SetConsoleMode(handle, new_mode):
            return

        self._term_handle = handle
        self._orig_mode = old_mode

    def exit_raw(self) -> None:
        """Restore the terminal to its previous mode (reverts "enter_raw").

        Returns:
            None
        """
        if self._term_handle is not None and self._orig_mode is not None:
            ctypes.windll.kernel32.SetConsoleMode(self._term_handle, self._orig_mode)

        self._term_handle = None
        self._orig_mode = None

    def _read_win_console_event(self) -> bytes:
        """Peek and consume a single Win32 console input event and map it to bytes.

        Returns:
            bytes: Data representing key input (including ANSI sequences for special keys)
        """
        if self._term_handle is None:
            return b""

        rec = INPUT_RECORD()
        nread = wintypes.DWORD()
        # Peek the buffer to avoid blocking read
        ok = ctypes.windll.kernel32.PeekConsoleInputW(self._term_handle, ctypes.byref(rec), 1, ctypes.byref(nread))
        if not ok or nread.value == 0:
            return b""
        # Force the read to consume the buffer events
        ctypes.windll.kernel32.ReadConsoleInputW(self._term_handle, ctypes.byref(rec), 1, ctypes.byref(nread))

        if rec.EventType == WINDOW_BUFFER_SIZE_EVENT:
            if self._resize_cb is not None:
                size = rec.Event.WindowBufferSize.dwSize
                cols, rows = int(size.X), int(size.Y)
                self._resize_cb(cols, rows)
            return b""

        if rec.EventType == KEY_EVENT:
            kev = rec.Event.KeyEvent
            if not kev.bKeyDown:
                return b""

            # Check if it is a special key
            ansi = KEYCODES.get(int(kev.wVirtualKeyCode))
            if ansi:
                return f"{ANSI_ESC}{ansi}".encode("utf-8")

            # Otherwise return the character
            ch = kev.uChar
            if ch is None:
                return b""
            try:
                return ch.encode("utf-8")
            except Exception:
                return b""

        return b""

    def install_input_reader(self, loop: Any, on_bytes: Callable[[bytes], None], on_close: Callable[[], None]) -> None:
        """Start reading from stdin and route incoming bytes to callbacks (connected to the manager handler).

        Args:
            loop (Any): The asyncio event loop instance used to register readers/tasks.
            on_bytes (Callable[[bytes], None]): Callback to handle received input bytes.
            on_close (Callable[[], None]): Callback to handle stdin closing or errors.

        Returns:
            None
        """
        if self._input_task is not None:
            return

        async def _pump() -> None:
            while True:
                try:
                    data = await loop.run_in_executor(None, self._read_win_console_event)
                except Exception:
                    on_close()
                    break

                if not data:
                    await asyncio.sleep(0.01)
                    continue

                try:
                    on_bytes(data)
                except Exception:
                    on_close()
                    break

        self._input_task = loop.create_task(_pump())

    def remove_input_reader(self, loop: Any) -> None:
        """Stop reading stdin previously configured by "install_input_reader".

         Args:
             loop (Any): The asyncio event loop instance used to deregister.

         Returns:
             None
         """
        if self._input_task is None:
            return

        self._input_task.cancel()
        self._input_task = None

    def write_stdout(self, data: bytes) -> None:
        """Write bytes to stdout.

        Args:
            data (bytes): Bytes to write to stdout.

        Returns:
            None
        """
        if self._stdout_handle is None:
            return

        text = data.decode("utf-8", errors="replace")
        buffer = ctypes.create_unicode_buffer(text)
        written = wintypes.DWORD(0)
        ok = ctypes.windll.kernel32.WriteConsoleW(self._stdout_handle, buffer, len(buffer), ctypes.byref(written), None)
        if not ok:
            raise RuntimeError

    def watch_resize(self, loop: Any, cb: Callable[[int, int], None]) -> None:
        """Start monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to register signal handlers or callbacks.
            cb (Callable[[int, int], None]): Callback to handle the resize.

        Returns:
            None

        Notes:
            Unused, resize is read from console events, only installs the callback.
        """
        self._resize_cb = cb

    def unwatch_resize(self, loop: Any) -> None:
        """Stop monitoring terminal resize events.

        Args:
            loop (Any): The asyncio event loop instance used to deregister signal handlers or callbacks.

        Returns:
            None

        Notes:
            Unused, resize is read from console events.
        """
        pass



from abc import ABC, abstractmethod
from typing import Any, Optional


class ITerminalSession(ABC):
    """Abstract interface representing an interactive terminal session.
    Instances should wrap a manager session (bidirectional byte stream) with resize support.

    Attributes:
        _handler (Any): A manager-specific handler of the session.
        _client (Optional[Any]): Manager client instance.
        _closed (bool): Boolean flag indicating whether session termination was requested.
    """

    __slots__ = ['_handler', '_client', '_closed']

    def __init__(self, handler: Any, client: Any) -> None:
        self._handler: Any = handler
        self._client: Optional[Any] = client

        self._closed: bool = False

    def fileno(self) -> Optional[int]:
        """Return an OS-level file descriptor for the session, if available.

        Returns:
            Optional[int]: The file descriptor, or None if the session cannot be polled via fd-based readiness APIs.
        """
        raise NotImplementedError

    @abstractmethod
    def read(self, n: int = 4096) -> bytes:
        """Read up to n bytes from the session output stream.

        Args:
            n (int): Maximum number of bytes to read.

        Returns:
            bytes: Data read from the session.
        """
        raise NotImplementedError("You must implement `read` method.")

    @abstractmethod
    def write(self, data: bytes) -> None:
        """Write bytes to the session input stream.

        Args:
            data (bytes): Data to send to the session.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `write` method.")

    @abstractmethod
    def resize(self, cols: int, rows: int) -> None:
        """Resize the session terminal dimensions.

        Args:
            cols (int): Terminal width in columns.
            rows (int): Terminal height in rows.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `resize` method.")

    @abstractmethod
    def close(self) -> None:
        """Close the session and release resources.

        Returns:
            None
        """
        raise NotImplementedError("You must implement `close` method.")



import collections
import logging
from itertools import chain
from typing import Dict, Set, Any, List, Union, Optional, Tuple

from fs import open_fs
from fs.base import FS

from . import Interface as InterfacePackage
from . import Link as LinkPackage
from . import Machine as MachinePackage
from .ExternalLink import ExternalLink
from .. import utils
from ..exceptions import LinkNotFoundError, MachineNotFoundError, MachineAlreadyExistsError, LinkAlreadyExistsError, \
    InvocationError
from ..foundation.model.LabFilesystemMixin import LabFilesystemMixin

LAB_METADATA: List[str] = ["LAB_NAME", "LAB_DESCRIPTION", "LAB_VERSION", "LAB_AUTHOR", "LAB_EMAIL", "LAB_WEB"]


class Lab(LabFilesystemMixin):
    """A Kathara network scenario, containing information about devices and collision domains.

    Attributes:
        name (str): The name of the network scenario.
        description (str): A short description of the network scenario.
        version (str): The version of the network scenario.
        author (str): The author of the network scenario.
        email (str): The email of the author of the network scenario.
        web (str): The web address of the author of the network scenario.
        hash (str): The hash identifier of the network scenario.
        machines (Dict[str, Kathara.model.Machine]): The devices of the network scenario. Keys are device names, Values
            are Kathara device objects.
        links (Dict[str, Kathara.model.Link]): The collision domains of the network scenario.
            Keys are collision domains names, Values are Kathara collision domain objects.
        general_options (Dict[str, Any]): The general options of the network scenario.
        global_machine_metadata (Dict[str, Any]): Metadata to apply to all the devices of the network scenario at
            the startup.
        has_dependencies (bool): True if there are dependencies among the devices boot.
        shared_path (str): Path to shared folder of the network scenario, if the network scenario has a real OS path.
        fs (fs.FS): The filesystem of the network scenario. Contains files and configurations associated to it.
    """
    __slots__ = ['_name', 'description', 'version', 'author', 'email', 'web', 'hash',
                 'machines', 'links', 'general_options', 'global_machine_metadata', 'has_dependencies', 'shared_path']

    def __init__(self, name: Optional[str], path: Optional[str] = None) -> None:
        """Create a new instance of a Kathara network scenario.

        Args:
            name (str): The name of the network scenario.
            path (str): The path to the network scenario directory, if exists.

        Returns:
            None
        """
        super().__init__()

        self._name: Optional[str] = name
        self.description: Optional[str] = None
        self.version: Optional[str] = None
        self.author: Optional[str] = None
        self.email: Optional[str] = None
        self.web: Optional[str] = None

        self.machines: Dict[str, 'MachinePackage.Machine'] = {}
        self.links: Dict[str, 'LinkPackage.Link'] = {}

        self.general_options: Dict[str, Any] = {}

        self.global_machine_metadata: Dict[str, Any] = {}

        self.has_dependencies: bool = False

        self.shared_path: Optional[str] = None

        self.hash: str = utils.generate_urlsafe_hash(path if self._name is None else self._name)

        if path:
            self.fs: FS = open_fs(f"osfs://{path}")
        else:
            self.fs: FS = open_fs("mem://")

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value
        self.hash = utils.generate_urlsafe_hash(value)

    def connect_machine_to_link(self, machine_name: str, link_name: str,
                                machine_iface_number: int = None, mac_address: Optional[str] = None) \
            -> Tuple['MachinePackage.Machine', 'InterfacePackage.Interface']:
        """Connect the specified device to the specified collision domain.

        Args:
            machine_name (str): The device name.
            link_name (str): The collision domain name.
            machine_iface_number (int): The number of the device interface to connect. If it is None, the first free
                number is used.
            mac_address (Optional[str]): The MAC address to assign to the interface.

        Returns:
            Tuple[Kathara.model.Machine.Machine, Kathara.model.Interface.Interface]: A tuple containing the Kathara
                device and the interface object specified by their names.

        Raises:
            Exception: If an already used interface number is specified.
        """
        machine = self.get_or_new_machine(machine_name)
        link = self.get_or_new_link(link_name)

        interface = machine.add_interface(link, number=machine_iface_number, mac_address=mac_address)

        return machine, interface

    def connect_machine_obj_to_link(self, machine: 'MachinePackage.Machine', link_name: str,
                                    machine_iface_number: int = None, mac_address: Optional[str] = None) \
            -> 'InterfacePackage.Interface':
        """Connect the specified device object to the specified collision domain.

        Args:
            machine (Kathara.model.Machine): The device object.
            link_name (str): The collision domain name.
            machine_iface_number (int): The number of the device interface to connect. If it is None, the first free
                number is used.
            mac_address (Optional[str]): The MAC address to assign to the interface.

        Returns:
            Kathara.model.Interface.Interface: The interface object associated to the new interface..

        Raises:
            Exception: If an already used interface number is specified.
        """
        link = self.get_or_new_link(link_name)

        interface = machine.add_interface(link, number=machine_iface_number, mac_address=mac_address)

        return interface

    def assign_meta_to_machine(self, machine_name: str, meta_name: str, meta_value: str) -> Optional[Any]:
        """Assign meta information to the specified device.

        Args:
            machine_name (str): The name of the device.
            meta_name (str): The name of the meta property.
            meta_value (str): The value of the meta property.

        Returns:
            Optional[Any]: Previous value if meta was already assigned, None otherwise.

        Raises:
            MachineOptionError: If invalid values are specified for meta properties.
        """
        machine = self.get_or_new_machine(machine_name)

        return machine.add_meta(meta_name, meta_value)

    def attach_external_links(self, external_links: Dict[str, List[ExternalLink]]) -> None:
        """Attach external collision domains to the network scenario.

        Args:
            external_links (Dict[Kathara.model.Link, List[Kathara.model.ExternalLink]]): Keys are Link objects,
            values are ExternalLink objects.

        Returns:
            None

        Raises:
            LinkNotFoundError: If the external collision domain specified is not associated to the network scenario.
        """
        for (link_name, link_external_links) in external_links.items():
            if link_name not in self.links:
                raise LinkNotFoundError("Collision domain `%s` (declared in lab.ext) not found in network scenario "
                                        "collision domains." % link_name)

            self.links[link_name].external += link_external_links

    def check_integrity(self) -> None:
        """Check if the network interfaces numbers of all the devices in the network scenario are correctly assigned.

        Returns:
            None

        Raises:
            NonSequentialMachineInterfaceError: If there is a missing interface number in any device of the lab.
        """
        logging.debug("Checking network scenario integrity...")

        for machine in self.machines.values():
            machine.check()

    def get_links_from_machines(self, machines: Union[List[str], Set[str]]) -> Set[str]:
        """Return the name of the collision domains connected to the devices.

        Args:
            machines (Union[List[str], Set[str]]): A set or a list with selected devices names.

        Returns:
            Set[str]: A set of names of collision domains.
        """
        # Intersect selected machines names with self.machines keys
        machines = set(self.machines.keys()) & set(machines)
        # Apply filtering
        machines = [v for (k, v) in self.machines.items() if k in machines]

        # Get only selected machines Link objects.
        selected_links = set(chain.from_iterable([machine.interfaces.values() for machine in machines]))
        selected_links = {interface.link.name for interface in selected_links}

        return selected_links

    def get_links_from_machine_objs(self,
                                    machines: Union[List['MachinePackage.Machine'], Set['MachinePackage.Machine']]) -> \
            Set[str]:
        """Return the name of the collision domains connected to the devices.

        Args:
            machines (Union[List[str], Set[str]]): A set or a list with selected devices names.

        Returns:
            Set[str]: A set of names of collision domains.
        """
        # Intersect selected machines names with self.machines keys
        machines = set(self.machines.keys()) & set(map(lambda x: x.name, machines))
        # Apply filtering
        machines = [v for (k, v) in self.machines.items() if k in machines]

        # Get only selected machines Link objects.
        selected_links = set(chain.from_iterable([machine.interfaces.values() for machine in machines]))
        selected_links = {interface.link.name for interface in selected_links}

        return selected_links

    def apply_dependencies(self, dependencies: List[str]) -> None:
        """Order the list of devices of the network scenario to satisfy the boot dependencies.

        Args:
            dependencies (List[str]): If not empty, dependencies are applied.

        Returns:
            None
        """

        def dep_sort(item: str) -> int:
            try:
                return dependencies.index(item) + 1
            except ValueError:
                return 0

        self.machines = collections.OrderedDict(sorted(self.machines.items(), key=lambda t: dep_sort(t[0])))
        self.has_dependencies = True

    def get_machine(self, name: str) -> 'MachinePackage.Machine':
        """Get the specified device.

        Args:
            name (str): The name of the device

        Returns:
            Kathara.model.Machine: A Kathara device.

        Raises:
            MachineNotFoundError: If the device is not in the network scenario.
        """
        if name not in self.machines:
            raise MachineNotFoundError(f"Device {name} not in the network scenario.")

        return self.machines[name]

    def new_machine(self, name: str, **kwargs) -> 'MachinePackage.Machine':
        """Create and add the device to the devices list.

        Args:
            name (str): The name of the device
            **kwargs: Contains device meta information.
                Keys are meta property names, values are meta property values.

        Returns:
            Kathara.model.Machine: A Kathara device.

        Raises:
            MachineAlreadyExistsError: If the device is already in the network scenario.
        """
        if name in self.machines:
            raise MachineAlreadyExistsError(name)

        self.machines[name] = MachinePackage.Machine(self, name, **kwargs)

        return self.machines[name]

    def get_or_new_machine(self, name: str, **kwargs) -> 'MachinePackage.Machine':
        """Get the specified device. If it not exists, create and add it to the devices list.

        Args:
            name (str): The name of the device
            **kwargs: Contains device meta information.
                Keys are meta property names, values are meta property values.

        Returns:
            Kathara.model.Machine: A Kathara device.
        """
        if name not in self.machines:
            self.machines[name] = MachinePackage.Machine(self, name, **kwargs)

        return self.machines[name]

    def remove_machine(self, name: Optional[str] = None, machine: Optional['MachinePackage.Machine'] = None,
                       delete_fs: bool = False) -> None:
        """Remove the device from the network scenario.

        Args:
            name (str): The name of the device to remove. Can be used as an alternative to machine.
                If None, machine should be set.
            machine (Kathara.model.Machine): The device to remove. Can be used as an alternative to name.
                If None, name should be set.
            delete_fs (bool): If true, the files and folders associated to the device will be removed from fs.
                WARNING: this option will delete the files and folders from the host if the fs is an `os` filesystem!

        Raises:
            InvocationError: If a device name or object is not specified.
            MachineNotFoundError: If the device is not found in the network scenario.
        """
        if name is None and machine is None:
            raise InvocationError("You must specify a device name or object.")

        if name is None:
            name = machine.name

        if name not in self.machines:
            raise MachineNotFoundError(f"Device {name} not in the network scenario.")

        for interface in self.machines[name].interfaces.values():
            del interface.link.machines[name]

        del self.machines[name]
        if delete_fs:
            if self.fs.exists(f"{name}.startup"):
                self.fs.remove(f"{name}.startup")

            if self.fs.exists(f"{name}.shutdown"):
                self.fs.remove(f"{name}.shutdown")

            if self.fs.exists(name):
                self.fs.removedir(name)

    def get_link(self, name: str) -> 'LinkPackage.Link':
        """Get the specified collision domain.

        Args:
            name (str): The name of the collision domain.

        Returns:
            Kathara.model.Link: A Kathara collision domain.

        Raises:
            LinkNotFoundError: If the specified link is not in the network scenario.
        """
        if name not in self.links:
            raise LinkNotFoundError(f"Collision domain {name} not found in the network scenario.")

        return self.links[name]

    def new_link(self, name: str) -> 'LinkPackage.Link':
        """Create the collision domain and add it to the collision domains list.

        Args:
            name (str): The name of the collision domain.

        Returns:
            Kathara.model.Link: A Kathara collision domain.

        Raises:
            LinkAlreadyExistsError: If the specified link is already in the network scenario.
        """
        if name in self.links:
            raise LinkAlreadyExistsError(f"Collision domain {name} is already the network scenario.")

        self.links[name] = LinkPackage.Link(self, name)

        return self.links[name]

    def get_or_new_link(self, name: str) -> 'LinkPackage.Link':
        """Get the specified collision domain. If it not exists, create and add it to the collision domains list.

        Args:
            name (str): The name of the collision domain.

        Returns:
            Kathara.model.Link: A Kathara collision domain.
        """
        if name not in self.links:
            self.links[name] = LinkPackage.Link(self, name)

        return self.links[name]

    def create_shared_folder(self) -> None:
        """If the network scenario has a directory, create the network scenario shared folder.

        Returns:
            None

        Raises:
            IOError: If the shared folder is a Symlink, delete it.
            OSError: If there is a permission error.
        """
        if not self.has_host_path():
            return

        try:
            self.shared_path = self.fs.makedir('shared', recreate=True).getsyspath("")
            if self.fs.islink('shared'):
                raise ValueError("`shared` folder is a symlink, delete it.")
        except OSError:
            # Do not create shared folder if not permitted.
            return

    def has_host_path(self) -> bool:
        """Check if the network scenario has a directory on the host.

        Returns:
            bool: True if network scenario has a path on the host filesystem, else False.
        """
        return self.fs_type() == "os"

    def add_option(self, name: str, value: Any) -> None:
        """Add an option to the network scenario.

        Args:
            name (str): The name of the option.
            value (Any): The value of the option.

        Returns:
            None
        """
        if value is not None:
            self.general_options[name] = value

    def add_global_machine_metadata(self, name: str, value: Any) -> None:
        """Add a global machine metadata to the network scenario.

        Args:
            name (str): The name of the meta.
            value (Any): The value of the meta.

        Returns:
            None
        """
        if value is not None:
            self.global_machine_metadata[name] = value

    def has_machine(self, machine_name: str) -> bool:
        """Check if the specified device is in the network scenario.

        Args:
            machine_name (str): The name of the device to search.

        Returns:
            bool: True if the device is in the network scenario, else False.
        """
        return machine_name in self.machines.keys()

    def has_machines(self, machine_names: Set[str]) -> bool:
        """Check if the specified devices are in the network scenario.

        Args:
            machine_names (Set[str]): A set of strings containing the names of the devices to search.

        Returns:
            bool: True if the devices are all in the network scenario, else False.
        """
        return all(map(lambda x: self.has_machine(x), machine_names))

    def has_link(self, link_name: str) -> bool:
        """Check if the specified collision domain is in the network scenario.

        Args:
            link_name (str): The name of the collision domain to search.

        Returns:
            bool: True if the collision domain is in the network scenario, else False.
        """
        return link_name in self.links.keys()

    def has_links(self, link_names: Set[str]) -> bool:
        """Check if the specified collision domains are in the network scenario.

        Args:
            link_names (Set[str]): A set of strings containing the names of the collision domains to search.

        Returns:
            bool: True if the collision domains are all in the network scenario, else False.
        """
        return all(map(lambda x: self.has_link(x), link_names))

    def __repr__(self) -> str:
        return "Lab(%s, %s, %s, %s)" % (self.fs, self.hash, self.machines, self.links)

    def __str__(self) -> str:
        lab_info = []

        if self._name:
            lab_info.append(f"Name: {self._name}")
        if self.description:
            lab_info.append(f"Description: {self.description}")
        if self.version:
            lab_info.append(f"Version: {self.version}")
        if self.author:
            lab_info.append(f"Author(s): {self.author}")
        if self.email:
            lab_info.append(f"Email: {self.email}")
        if self.web:
            lab_info.append(f"Website: {self.web}")

        return "\n".join(lab_info)
